"""
Mentions for sources that do not support webmentions.
Needs to be input manually.
"""
from mentions.models.base import MentionsBaseModel
from mentions.models.mixins import QuotableMixin

__all__ = [
    "SimpleMention",
]


class SimpleMention(QuotableMixin, MentionsBaseModel):
    """A 'fake' webmention, created manually by you.

    If you find your content mentioned somewhere that did not submit a Webmention
    you may manually create a SimpleMention to represent it instead.

    SimpleMentions are returned alongside actual Webmentions by the `/get` endpoint
    so you can display them as you would any other Webmention."""

    pass



import json
from functools import reduce
from typing import List, Optional

import mf2py
from bs4 import Tag

from mentions.exceptions import NotEnoughData
from mentions.microformats import H_CARD, H_ENTRY, H_FEED
from mentions.models import HCard
from mentions.models.hcard import update_or_create_hcard

__all__ = [
    "parse_hcard",
    "find_related_hcard",
]


# Key names for mf2py parsing output
AUTHOR = "author"
CHILDREN = "children"
TYPE = "type"
PROPERTIES = "properties"
NAME = "name"
URL = "url"
PHOTO = "photo"


def parse_hcard(
    soup: Tag,
    recursive: bool = False,
) -> Optional[HCard]:
    """Create or update HCard(s) using data from a BeautifulSoup document.

    See https://github.com/microformats/mf2py"""

    parser = mf2py.Parser(doc=soup)
    parsed_data = parser.to_dict()
    items = parsed_data.get("items", [])
    return _find_hcard(items, recursive=recursive)


def find_related_hcard(link: Tag) -> Optional[HCard]:
    """Try to find a post-specific h-card from a parent `h-entry` or `h-feed`."""
    hentry = link.find_parent(class_=H_ENTRY)
    if hentry:
        hcard = parse_hcard(hentry, recursive=True)
        if hcard:
            return hcard

    hfeed = link.find_parent(class_=H_FEED)
    if hfeed:
        hcard = parse_hcard(hfeed, recursive=True)
        if hcard:
            return hcard


def _find_hcard(data: List[dict], recursive: bool = False) -> Optional[HCard]:
    """Find a useful `h-card` in parsed microformats data.

    Args:
        data: Parsed microformat data
        recursive: If True, traverse h-feed and h-entry containers to try
                   find embedded h-card.
                   If False, h-card will only be found at the top level.
    """

    fallback = []  # List of items that may contain an embedded h-card

    for item in data:
        _type = item.get(TYPE, [])

        if H_CARD in _type:
            try:
                hcard = _create_hcard(item)
                if hcard:
                    return hcard

            except NotEnoughData:
                # This h-card is missing required fields, keep looking for a better-formed one.
                pass

        elif recursive and (H_ENTRY in _type or H_FEED in _type):
            fallback.append(item)

    if not recursive:
        return None

    return _find_embedded_hcard(fallback)


def _find_embedded_hcard(items: List[dict]) -> Optional[HCard]:
    """Traverse `h-entry` and `h-feed` containers to find an `h-card`"""

    if not items:
        return None

    for item in items:
        _type = item.get(TYPE)
        props = item.get(PROPERTIES)

        if AUTHOR in props:
            return _find_hcard(props.get(AUTHOR))

        elif CHILDREN in item:
            return _find_hcard(item.get(CHILDREN, []))


def _create_hcard(data: dict) -> HCard:
    props = data.get(PROPERTIES)
    homepage = props.get(URL, [None])[0]
    name = props.get(NAME, [""])[0]
    avatar = props.get(PHOTO, [""])[0]
    _json = json.dumps(data, sort_keys=True)

    _require_any_of([name, homepage])

    return update_or_create_hcard(
        homepage=homepage,
        name=name,
        avatar=avatar,
        data=_json,
    )


def _require_any_of(fields: List[str]):
    has_required_fields = (
        reduce(lambda acc, value: acc + 1 if value else acc, fields, 0) >= 1
    )

    if not has_required_fields:
        raise NotEnoughData(
            "An HCard requires a value for at least one of 'name', 'homepage'"
        )



"""Options for `django-wm` are configured by adding attributes to Django settings.

These may be defined in either of the following two formats. Choose one format
or the other. If `WEBMENTIONS` is defined, the latter format will be ignored.

# settings.py
WEBMENTIONS = {
    "TIMEOUT": 10,
    "USE_CELERY": True,
}

--- or ---

# settings.py
WEBMENTIONS_TIMEOUT = 10
WEBMENTIONS_USE_CELERY = True

"""

import logging
from typing import Dict

from django.conf import settings

import mentions

__all__ = [
    "auto_approve",
    "dashboard_public",
    "domain_name",
    "get_config",
    "max_retries",
    "retry_interval",
    "target_requires_model",
    "timeout",
    "url_scheme",
    "use_celery",
    "user_agent",
]


NAMESPACE = "WEBMENTIONS"
SETTING_ALLOW_OUTGOING_DEFAULT = f"{NAMESPACE}_ALLOW_OUTGOING_DEFAULT"
SETTING_ALLOW_SELF_MENTIONS = f"{NAMESPACE}_ALLOW_SELF_MENTIONS"
SETTING_AUTO_APPROVE = f"{NAMESPACE}_AUTO_APPROVE"
SETTING_DASHBOARD_PUBLIC = f"{NAMESPACE}_DASHBOARD_PUBLIC"
SETTING_DEFAULT_URL_PARAMETER_MAPPING = f"{NAMESPACE}_DEFAULT_URL_PARAMETER_MAPPING"
SETTING_INCOMING_TARGET_MODEL_REQUIRED = f"{NAMESPACE}_INCOMING_TARGET_MODEL_REQUIRED"
SETTING_MAX_RETRIES = f"{NAMESPACE}_MAX_RETRIES"
SETTING_RETRY_INTERVAL = f"{NAMESPACE}_RETRY_INTERVAL"
SETTING_TIMEOUT = f"{NAMESPACE}_TIMEOUT"
SETTING_URL_SCHEME = f"{NAMESPACE}_URL_SCHEME"
SETTING_USE_CELERY = f"{NAMESPACE}_USE_CELERY"
SETTING_USER_AGENT = f"{NAMESPACE}_USER_AGENT"

"""settings.DOMAIN_NAME is sometimes used by other libraries for the same purpose,
no need to lock it to our namespace."""
SETTING_DOMAIN_NAME = "DOMAIN_NAME"

DEFAULTS = {
    SETTING_ALLOW_OUTGOING_DEFAULT: False,
    SETTING_ALLOW_SELF_MENTIONS: True,
    SETTING_AUTO_APPROVE: False,
    SETTING_DASHBOARD_PUBLIC: False,
    SETTING_DEFAULT_URL_PARAMETER_MAPPING: {"object_id": "id"},
    SETTING_DOMAIN_NAME: None,
    SETTING_INCOMING_TARGET_MODEL_REQUIRED: False,
    SETTING_MAX_RETRIES: 5,
    SETTING_RETRY_INTERVAL: 60 * 10,
    SETTING_TIMEOUT: 10,
    SETTING_URL_SCHEME: "https",
    SETTING_USE_CELERY: True,
    SETTING_USER_AGENT: f"django-wm/{mentions.__version__} (+{mentions.__url__})",
}


log = logging.getLogger(__name__)


def _get_attr(key: str):
    if not key.startswith(NAMESPACE) or not hasattr(settings, NAMESPACE):
        return getattr(settings, key, DEFAULTS[key])

    opts: dict = getattr(settings, NAMESPACE)
    simple_key = key[len(f"{NAMESPACE}_") :]
    return opts.get(simple_key, DEFAULTS[key])


def get_config() -> dict:
    return {key: _get_attr(key) for key in DEFAULTS.keys()}


def allow_outgoing_default() -> bool:
    """Return settings.WEBMENTIONS_ALLOW_OUTGOING_DEFAULT.

    Sets the default value for `MentionableMixin.allow_outgoing_webmentions`."""

    return _get_attr(SETTING_ALLOW_OUTGOING_DEFAULT)


def allow_self_mentions() -> bool:
    """Return settings.WEBMENTIONS_ALLOW_SELF_MENTIONS.

    If True, you can send webmentions to yourself.
    If False, outgoing links that target your own domain name will be ignored.
    """
    return _get_attr(SETTING_ALLOW_SELF_MENTIONS)


def auto_approve() -> bool:
    """Return settings.WEBMENTIONS_AUTO_APPROVE, or False if not set.

    If True, any received Webmentions will immediately become 'public': they will be included in `/get` API responses.
    If False, received Webmentions must be approved by a user with `approve_webmention` permissions.
    """
    return _get_attr(SETTING_AUTO_APPROVE)


def dashboard_public() -> bool:
    """Return settings.WEBMENTIONS_DASHBOARD_PUBLIC.

    This is intended to help with debugging while developing the `django-wm`
    library and probably should not be used otherwise."""
    is_dashboard_public = _get_attr(SETTING_DASHBOARD_PUBLIC)
    if not settings.DEBUG and is_dashboard_public:
        log.warning(
            f"settings.{SETTING_DASHBOARD_PUBLIC} should not be `True` when in production!"
        )

    return is_dashboard_public


def default_url_parameter_mapping() -> Dict[str, str]:
    """Return settings.WEBMENTIONS_DEFAULT_URL_PARAMETER_MAPPING.

    This is used by `MentionableMixin.resolve_from_url_kwargs` if you do not
    override it on your model.

    The first value is the name of the parameter captured from your URL pattern,
    the second is the name of the field on the model.

    e.g. The default {"object_id": "id"} will result in a model query that looks
    like `MyModel.objects.get(id=url_kwargs.get("object_id"))`.
    """
    return _get_attr(SETTING_DEFAULT_URL_PARAMETER_MAPPING)


def domain_name() -> str:
    """Return settings.DOMAIN_NAME."""
    return _get_attr(SETTING_DOMAIN_NAME)


def max_retries() -> int:
    """Return settings.WEBMENTIONS_MAX_RETRIES.

    We may retry processing of webmentions if the remote server is inaccessible.
    This specifies the maximum number of retries before we give up."""
    return _get_attr(SETTING_MAX_RETRIES)


def retry_interval() -> int:
    """Return settings.WEBMENTIONS_RETRY_INTERVAL.

    We may retry processing of webmentions if the remote server is inaccessible.
    This specifies the delay (in seconds) between attempts.

    If using `celery`, this should be precise. Otherwise, this will be treated
    as a minimum interval (and will also depend on `cron` or whatever you are
    using to schedule mention processing).

    Warning: If using RabbitMQ, it may need to be reconfigured if you want to
    specify an interval of more than 15 minutes. See
    https://docs.celeryq.dev/en/stable/userguide/calling.html#eta-and-countdown
    for more information."""
    return _get_attr(SETTING_RETRY_INTERVAL)


def target_requires_model() -> bool:
    """Return settings.WEBMENTIONS_INCOMING_TARGET_MODEL_REQUIRED.

    If True, incoming webmentions will only be accepted if they target a path
    that resolves to a `MentionableMixin` instance.

    If False, any target path can receive webmentions."""
    return _get_attr(SETTING_INCOMING_TARGET_MODEL_REQUIRED)


def timeout() -> float:
    """Return settings.WEBMENTIONS_TIMEOUT.

    Timeout (in seconds) used for network requests when sending or verifying webmentions."""
    return _get_attr(SETTING_TIMEOUT)


def url_scheme() -> str:
    """Return settings.WEBMENTIONS_URL_SCHEME.

    This defaults to `https` which is hopefully what your server is using.
    It's handy to be able to choose when debugging stuff though."""
    return _get_attr(SETTING_URL_SCHEME)


def use_celery() -> bool:
    """Return settings.WEBMENTIONS_USE_CELERY, or True if not set.

    This setting enables/disables the use of `celery` for running tasks.
    If disabled, user must run these tasks using `manage.py pending_mentions` management command."""
    return _get_attr(SETTING_USE_CELERY)


def user_agent() -> str:
    """Return settings.WEBMENTIONS_USER_AGENT.

    This is included with all network requests made by `django-wm`."""
    return _get_attr(SETTING_USER_AGENT)



from enum import Enum
from typing import List, Tuple

from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from mentions import microformats

__all__ = [
    "IncomingMentionType",
    "QuotableMixin",
]


class IncomingMentionType(Enum):
    """Properties that describe the context of the incoming webmention.

    See: https://microformats.org/wiki/h-entry"""

    Bookmark = microformats.BOOKMARK
    Like = microformats.LIKE
    Listen = microformats.LISTEN
    Reply = microformats.REPLY
    Repost = microformats.REPOST
    Translation = microformats.TRANSLATION
    Watch = microformats.WATCH

    @classmethod
    def choices(cls) -> List[Tuple[str, str]]:
        return [(x.lower(), x) for x in cls.__members__.keys()]

    @classmethod
    def serialized_names(cls) -> List[str]:
        return [x.lower() for x in cls.__members__.keys()]

    @classmethod
    def get_microformat_from_name(cls, name: str) -> str:
        """Get the microformat class name for the given mention type."""
        return cls[name.capitalize()].get_microformat() if name else ""

    def get_microformat(self) -> str:
        """Get the microformat class name for this mention type."""
        return self.value

    def serialized_name(self) -> str:
        """The value used in the database and API responses."""
        return self.name.lower()


class QuotableMixin(models.Model):
    """A third-party thing that talks about a first-party thing."""

    class Meta:
        abstract = True

    target_url = models.URLField(
        _("target URL"),
        help_text=_("Our URL that is mentioned."),
    )
    source_url = models.URLField(
        _("source URL"),
        help_text=_("The URL that mentions our content."),
    )

    quote = models.CharField(
        _("quote"),
        max_length=300,
        blank=True,
        null=True,
        help_text=_("A short excerpt from the quoted piece."),
    )

    post_type = models.CharField(
        _("post type"),
        max_length=64,
        null=True,
        blank=True,
        help_text=_("Type (e.g. like, reply, etc.) of mention, if specified."),
        choices=IncomingMentionType.choices(),
    )

    published = models.DateTimeField(
        _("published"),
        default=timezone.now,
    )

    hcard = models.ForeignKey(
        "HCard",
        verbose_name=_("h-card"),
        blank=True,
        null=True,
        on_delete=models.CASCADE,
    )

    # The mention target may be an instance of any class that
    # uses the MentionableMixin so we need to use a GenericForeignKey
    content_type = models.ForeignKey(
        ContentType,
        null=True,
        on_delete=models.CASCADE,
    )
    object_id = models.PositiveIntegerField(
        _("object ID"),
        blank=True,
        null=True,
    )
    target_object = GenericForeignKey(
        "content_type",
        "object_id",
    )

    def __str__(self):
        return (
            f"{_trim_to_length(self.source_url)} -> {_trim_to_length(self.target_url)}"
        )


def _trim_to_length(text: str, max_length=32) -> str:
    if len(text) >= max_length:
        return f"{text[:max_length - 1]}…"
    else:
        return text



import logging
import re
from typing import Optional

from bs4 import BeautifulSoup
from requests.structures import CaseInsensitiveDict

from mentions.util import html_parser

__all__ = [
    "get_endpoint_in_html",
    "get_endpoint_in_html_body",
    "get_endpoint_in_html_head",
    "get_endpoint_in_http_headers",
]

HTTP_LINK_PATTERN = re.compile(
    r"<(?P<url>[^>]+?)>;([^,]*;)*\s*rel=(?P<quote>['\"]?)webmention(?P=quote)"
)

log = logging.getLogger(__name__)


def get_endpoint_in_http_headers(headers: CaseInsensitiveDict) -> Optional[str]:
    link = headers.get("Link")
    if link is None:
        return

    if "webmention" in link:
        match = re.search(HTTP_LINK_PATTERN, link)
        if match is None:
            return

        endpoint = match.group("url")
        log.debug(f"Webmention endpoint found in HTTP header: '{endpoint}'")
        return endpoint


def get_endpoint_in_html(html: str) -> Optional[str]:
    """Search for a webmention endpoint in HTML."""
    soup = html_parser(html)

    return get_endpoint_in_html_head(soup) or get_endpoint_in_html_body(soup)


def get_endpoint_in_html_head(soup: BeautifulSoup) -> Optional[str]:
    """Check HTML <head> for <link> webmention endpoint."""
    links = soup.head.find_all("link", href=True, rel=True)
    for link in links:
        if "webmention" in link["rel"]:
            endpoint = link["href"]
            log.debug(f"Webmention endpoint found in document head: {endpoint}")
            return endpoint


def get_endpoint_in_html_body(soup: BeautifulSoup) -> Optional[str]:
    """Check HTML <body> for <a> webmention endpoint."""
    links = soup.body.find_all("a", href=True, rel=True)
    for link in links:
        if "webmention" in link["rel"]:
            endpoint = link["href"]
            log.debug(f"Webmention endpoint found in document body: {endpoint}")
            return endpoint



# Generated by Django 4.1.1 on 2022-09-27 18:09

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("mentions", "0008_alter_webmention_options"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="outgoingwebmentionstatus",
            options={
                "ordering": ["-created_at"],
                "verbose_name_plural": "Outgoing Webmentions",
            },
        ),
    ]



import logging

from mentions.exceptions import RejectedByConfig, SourceNotAccessible
from mentions.models import Webmention
from mentions.tasks.incoming.process import verify_webmention
from mentions.tasks.incoming.status import Status

log = logging.getLogger(__name__)


def reverify_mention(mention: Webmention) -> bool:
    """Redo the verification process for the given Webmention.

    Returns True if the mention was modified, False otherwise."""

    source_url = mention.source_url
    target_url = mention.target_url

    status = Status()

    try:
        is_verified, target_object, metadata = verify_webmention(
            source_url=source_url,
            target_url=target_url,
        )

    except RejectedByConfig as e:
        return _mark_invalid(mention, status.warning(str(e)))

    except SourceNotAccessible:
        return _mark_invalid(
            mention, status.warning(f"Source URL not accessible: '{source_url}'")
        )

    updated_fields = []

    if mention.target_object != target_object:
        log.info(f"Update target_object: {mention.target_object} -> {target_object}")
        mention.target_object = target_object
        updated_fields.append("content_type")
        updated_fields.append("object_id")

    if metadata:
        if mention.hcard != metadata.hcard:
            log.info(f"Update hcard: {mention.hcard} -> {metadata.hcard}")
            mention.hcard = metadata.hcard
            updated_fields.append("hcard")

        if mention.post_type != metadata.post_type:
            log.info(f"Update post_type: {mention.post_type} -> {metadata.post_type}")
            mention.post_type = metadata.post_type
            updated_fields.append("post_type")

    if mention.validated != is_verified:
        log.info(f"Update validated: {mention.validated} -> {is_verified}")
        mention.validated = is_verified
        updated_fields.append("validated")

    if updated_fields:
        updated_fields.append("notes")
        _append_notes(mention, status.info(f"Updated fields: {updated_fields}"))
        mention.save(update_fields=updated_fields)
        return True

    log.info("Webmention unchanged.")
    return False


def _mark_invalid(mention: Webmention, status: Status) -> bool:
    modified = mention.validated
    mention.validated = False
    _append_notes(mention, status)
    mention.save()
    return modified


def _append_notes(mention: Webmention, status: Status):
    mention.notes = "\n".join(
        [note for note in [mention.notes, *status.notes] if note]
    )[:1023]



from collections import namedtuple
from urllib.parse import urlsplit

from django.core.validators import URLValidator

__all__ = [
    "get_base_url",
    "get_domain",
    "get_urlpath",
    "get_url_validator",
]

UrlParts = namedtuple("UrlParts", ["scheme", "domain", "path", "query", "fragment"])


def get_url_validator() -> URLValidator:
    return URLValidator(schemes=["http", "https"])


def split_url(url: str) -> UrlParts:
    scheme, full_domain, path, query, fragment = urlsplit(url)
    domain = full_domain.split(":")[0]  # Remove port number if present
    return UrlParts(scheme, domain, path, query, fragment)


def get_base_url(url: str) -> str:
    scheme, domain, _, _, _ = split_url(url)
    return f"{scheme}://{domain}"


def get_domain(url: str) -> str:
    return split_url(url).domain


def get_urlpath(url: str) -> str:
    """Return the path component of the given URL."""

    return split_url(url).path



# Generated by Django 4.0.2 on 2022-02-03 20:45

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mentions', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='PendingIncomingWebmention',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('source_url', models.URLField(help_text='The URL of the content that mentions your content.')),
                ('target_url', models.URLField(help_text='The URL of the page on your server that is being mentioned.')),
                ('sent_by', models.URLField(help_text='The origin of the webmention request.')),
            ],
            options={
                'ordering': ['-created_at'],
            },
        ),
        migrations.CreateModel(
            name='PendingOutgoingContent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_at', models.DateTimeField(auto_now_add=True, null=True)),
                ('absolute_url', models.URLField(help_text='URL on our server where the content can be found.')),
                ('text', models.TextField(help_text='Text that may contain mentionable links. (retrieved via MentionableMixin.all_text())')),
            ],
            options={
                'ordering': ['-created_at'],
            },
        ),
    ]



import logging
from typing import Optional, Tuple
from urllib.parse import urljoin

from django.core.exceptions import ValidationError
from requests import RequestException, Response

from mentions import config
from mentions.exceptions import TargetNotAccessible
from mentions.models import OutgoingWebmentionStatus
from mentions.models.outgoing_status import get_or_create_outgoing_webmention
from mentions.tasks.outgoing.parsing import (
    get_endpoint_in_html,
    get_endpoint_in_http_headers,
)
from mentions.util import get_url_validator, http_get, http_post

__all__ = [
    "try_send_webmention",
]


STATUS_MESSAGE_TARGET_UNREACHABLE = "The target URL could not be retrieved: {error}."
STATUS_MESSAGE_TARGET_ERROR_CODE = (
    "The target URL returned an HTTP error code: {status_code}."
)
STATUS_MESSAGE_TARGET_ENDPOINT_ERROR = (
    "The target endpoint URL returned an HTTP error code"
)
STATUS_MESSAGE_OK = "The target server accepted the webmention."

log = logging.getLogger(__name__)


def try_send_webmention(
    source_urlpath: str,
    target_url: str,
    outgoing_status: Optional[OutgoingWebmentionStatus],
) -> Optional[bool]:
    """Try to send a webmention for target_url.

    Returns:
        True if a webmention was submitted successfully.
        False if a webmention endpoint was resolved but submission failed.
        None if a webmention endpoint could not be resolved (i.e. the website does not appear to support webmentions).
    """
    if outgoing_status is None:
        outgoing_status = get_or_create_outgoing_webmention(source_urlpath, target_url)

    try:
        response = _get_target(outgoing_status, target_url)
    except TargetNotAccessible:
        return

    endpoint = _get_absolute_endpoint_from_response(response)
    if endpoint:
        log.debug(f"Found webmention endpoint: '{endpoint}'")
        return _try_send_webmention(
            outgoing_status,
            source_urlpath=source_urlpath,
            endpoint=endpoint,
            target_url=target_url,
        )

    else:
        log.info(f"No webmention endpoint found for url '{target_url}'")


def _get_target(
    status: OutgoingWebmentionStatus,
    target_url: str,
) -> Optional[Response]:
    """Confirm the target is accessible."""
    log.debug(f"Checking url='{target_url}' for webmention support...")
    try:
        response = http_get(target_url)
        if response.status_code < 300:
            return response

        error_message = STATUS_MESSAGE_TARGET_ERROR_CODE.format(
            status_code=response.status_code
        )

    except RequestException as e:
        error_message = STATUS_MESSAGE_TARGET_UNREACHABLE.format(error=e)

    _save_for_retry(
        status,
        error_message,
    )
    raise TargetNotAccessible()


def _try_send_webmention(
    status: OutgoingWebmentionStatus,
    source_urlpath: str,
    endpoint: str,
    target_url: str,
):
    success, status_code = _send_webmention(source_urlpath, endpoint, target_url)

    status.target_webmention_endpoint = endpoint
    status.response_code = status_code

    if success:
        log.info(f"Webmention submission successful for '{target_url}'")
        status.status_message = STATUS_MESSAGE_OK
        status.successful = True
        status.mark_processing_successful(save=True)
        return True

    else:
        log.warning(
            f"Webmention submission failed for '{target_url}' [endpoint={endpoint}]"
        )
        _save_for_retry(status, STATUS_MESSAGE_TARGET_ENDPOINT_ERROR)
        return False


def _send_webmention(
    source_urlpath: str,
    endpoint: str,
    target: str,
) -> Tuple[bool, int]:
    payload = {
        "target": target,
        "source": config.build_url(source_urlpath),
    }
    response = http_post(endpoint, data=payload)
    status_code = response.status_code

    if status_code >= 300:
        log.warning(
            f'Sending webmention to "{endpoint}" '
            f"FAILED with status_code={status_code}"
        )
        return False, status_code

    else:
        log.info(
            f'Sending webmention to "{endpoint}" '
            f"successful with status_code={status_code}"
        )
        return True, status_code


def _get_absolute_endpoint_from_response(response: Response) -> Optional[str]:
    endpoint = get_endpoint_in_http_headers(response.headers) or get_endpoint_in_html(
        response.text
    )

    if endpoint:
        return _relative_to_absolute_url(response, endpoint)


def _relative_to_absolute_url(response: Response, url: str) -> Optional[str]:
    """
    If given url is relative, try to construct an absolute url using response domain.
    """
    try:
        get_url_validator()(url)
        return url  # url is already well-formed.
    except ValidationError:
        absolute_url = urljoin(response.url, url)
        log.debug(f"Relative endpoint url resolved: {url} -> {absolute_url}")
        return absolute_url


def _save_for_retry(status: OutgoingWebmentionStatus, message: str) -> None:
    """In case of network errors, mark the status for reprocessing later."""
    log.warning(message)
    status.status_message = message
    status.mark_processing_failed(save=True)



from django.db import models
from django.db.models import UniqueConstraint
from django.utils.translation import gettext_lazy as _

from mentions.models.base import MentionsBaseModel
from mentions.models.mixins import RetryableMixin

__all__ = [
    "PendingIncomingWebmention",
    "PendingOutgoingContent",
]


class PendingIncomingWebmention(RetryableMixin, MentionsBaseModel):
    """Temporary store of data about an incoming webmention.

    Only used if settings.WEBMENTIONS_USE_CELERY is False.
    Use `manage.py pending_mentions` to process."""

    source_url = models.URLField(
        _("source URL"),
        help_text=_("The URL of the content that mentions your content."),
    )
    target_url = models.URLField(
        _("target URL"),
        help_text=_("The URL of the page on your server that is being mentioned."),
    )
    sent_by = models.URLField(
        _("sent by"),
        help_text=_("The origin of the webmention request."),
    )

    def __str__(self):
        return f"PendingIncomingWebmention: {self.source_url} -> {self.target_url}"

    class Meta:
        constraints = [
            UniqueConstraint(
                fields=("source_url", "target_url"),
                name="unique_source_url_per_target_url",
            ),
        ]
        ordering = ["-created_at"]
        verbose_name = _("pending incoming webmention")
        verbose_name_plural = _("pending incoming webmentions")


class PendingOutgoingContent(MentionsBaseModel):
    """Temporary store of data about content that may contain outgoing webmentions.

    Use `manage.py pending_mentions` to process."""

    absolute_url = models.URLField(
        _("absolute URL"),
        help_text=_("URL on our server where the content can be found."),
        unique=True,
    )
    text = models.TextField(
        _("HTML content"),
        help_text=_("HTML text that may contain mentionable links."),
    )

    def __str__(self):
        return f"PendingOutgoingContent: {self.absolute_url}"

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("pending outgoing content")
        verbose_name_plural = _("pending outgoing content")



from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

from mentions import options

__all__ = [
    "RetryableMixin",
]


class RetryableMixin(models.Model):
    """A mixin for models that need to track reprocessing attempts."""

    class Meta:
        abstract = True

    retry_attempt_count = models.PositiveSmallIntegerField(
        _("retry attempt count"),
        default=0,
        editable=False,
        help_text=_("How many times we have attempted to process this object."),
    )
    last_retry_attempt = models.DateTimeField(
        _("latest retry attempt"),
        null=True,
        blank=True,
        help_text=_("When we last attempted to process this object."),
    )
    is_awaiting_retry = models.BooleanField(
        _("awaiting retry"),
        default=True,
        help_text=_(
            "Whether this object is eligible for another attempt at processing."
        ),
    )
    is_retry_successful = models.BooleanField(
        _("retry successful"),
        default=False,
        help_text=_("Whether this object has been processed successfully."),
        editable=False,
    )

    def mark_processing_failed(self, save: bool = False, now=timezone.now) -> None:
        if callable(now):
            now = now()

        self.retry_attempt_count += 1
        self.last_retry_attempt = now
        self.is_awaiting_retry = self.retry_attempt_count < options.max_retries()
        self.is_retry_successful = False

        if save:
            self.save()

    def mark_processing_successful(self, save: bool = False, now=timezone.now) -> None:
        if callable(now):
            now = now()

        self.retry_attempt_count += 1
        self.last_retry_attempt = now
        self.is_awaiting_retry = False
        self.is_retry_successful = True

        if save:
            self.save()

    def can_retry(self, now=timezone.now) -> bool:
        """Return True if awaiting_retry and enough time has passed since last_retry_attempt."""

        if self.is_retry_successful:
            return False

        if not self.is_awaiting_retry:
            return False

        if self.retry_attempt_count >= options.max_retries():
            return False

        if self.last_retry_attempt is None:
            return True

        if callable(now):
            now = now()

        seconds_since_last = (now - self.last_retry_attempt).total_seconds()
        return self.is_awaiting_retry and seconds_since_last >= options.retry_interval()

    def reset_retries(self):
        self.retry_attempt_count = 0
        self.is_awaiting_retry = True
        self.is_retry_successful = False
        self.save(
            update_fields=[
                "retry_attempt_count",
                "is_awaiting_retry",
                "is_retry_successful",
            ]
        )



# Generated by Django 4.0.1 on 2022-01-24 13:43

from django.db import migrations, models
import django.db.models.deletion
import mentions.models.webmention


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        ("contenttypes", "0002_remove_content_type_name"),
    ]

    operations = [
        migrations.CreateModel(
            name="HCard",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "name",
                    models.CharField(
                        blank=True,
                        help_text="Name of the person/organisation",
                        max_length=50,
                    ),
                ),
                (
                    "avatar",
                    models.URLField(
                        blank=True, help_text="Link to their profile image"
                    ),
                ),
                (
                    "homepage",
                    models.URLField(
                        blank=True, help_text="Link to their homepage", null=True
                    ),
                ),
                (
                    "json",
                    models.TextField(
                        blank=True, help_text="Raw json representation of this hcard"
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
        migrations.CreateModel(
            name="OutgoingWebmentionStatus",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "source_url",
                    models.URLField(
                        help_text="The URL on your server where this mention originates"
                    ),
                ),
                (
                    "target_url",
                    models.URLField(help_text="The URL that you mentioned."),
                ),
                (
                    "target_webmention_endpoint",
                    models.URLField(
                        blank=True,
                        help_text="The endpoint URL to which we sent the webmention",
                        null=True,
                    ),
                ),
                (
                    "status_message",
                    models.CharField(
                        help_text="Success, or an explanation of what went wrong.",
                        max_length=1024,
                    ),
                ),
                ("response_code", models.PositiveIntegerField(default=0)),
                ("successful", models.BooleanField(default=False)),
            ],
            options={
                "verbose_name_plural": "Outgoing Webmention Statuses",
                "ordering": ["-created_at"],
            },
        ),
        migrations.CreateModel(
            name="Webmention",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, null=True)),
                ("target_url", models.URLField(help_text="Our URL that is mentioned")),
                (
                    "source_url",
                    models.URLField(help_text="The URL that mentions our content"),
                ),
                (
                    "quote",
                    models.CharField(
                        blank=True,
                        help_text="A short excerpt from the quoted piece",
                        max_length=300,
                        null=True,
                    ),
                ),
                ("published", models.DateTimeField(auto_now_add=True)),
                ("object_id", models.PositiveIntegerField(blank=True, null=True)),
                (
                    "sent_by",
                    models.URLField(
                        blank=True,
                        help_text="Source address of the HTTP request that sent this webmention",
                    ),
                ),
                (
                    "approved",
                    models.BooleanField(
                        default=mentions.models.webmention._approve_default,
                        help_text="Allow this webmention to appear publicly",
                    ),
                ),
                (
                    "validated",
                    models.BooleanField(
                        default=False,
                        help_text="True if both source and target have been validated, confirmed to exist, and source really does link to target",
                    ),
                ),
                ("notes", models.CharField(blank=True, max_length=1024)),
                (
                    "content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "hcard",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="mentions.hcard",
                    ),
                ),
            ],
            options={
                "ordering": ["-created_at"],
                "permissions": [
                    (
                        "approve_webmention",
                        "Can approve received Webmentions for publishing.",
                    )
                ],
            },
        ),
        migrations.CreateModel(
            name="SimpleMention",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, null=True)),
                ("target_url", models.URLField(help_text="Our URL that is mentioned")),
                (
                    "source_url",
                    models.URLField(help_text="The URL that mentions our content"),
                ),
                (
                    "quote",
                    models.CharField(
                        blank=True,
                        help_text="A short excerpt from the quoted piece",
                        max_length=300,
                        null=True,
                    ),
                ),
                ("published", models.DateTimeField(auto_now_add=True)),
                ("object_id", models.PositiveIntegerField(blank=True, null=True)),
                (
                    "content_type",
                    models.ForeignKey(
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="contenttypes.contenttype",
                    ),
                ),
                (
                    "hcard",
                    models.ForeignKey(
                        blank=True,
                        null=True,
                        on_delete=django.db.models.deletion.CASCADE,
                        to="mentions.hcard",
                    ),
                ),
            ],
            options={
                "abstract": False,
            },
        ),
    ]



from django.core.management import BaseCommand

from mentions.tasks.scheduling import handle_pending_webmentions

PENDING_TYPE_ALL = "all"
PENDING_TYPE_INCOMING = "incoming"
PENDING_TYPE_OUTGOING = "outgoing"


class Command(BaseCommand):
    def add_arguments(self, parser):
        parser.add_argument(
            "pending_type",
            choices=[PENDING_TYPE_ALL, PENDING_TYPE_INCOMING, PENDING_TYPE_OUTGOING],
            default=PENDING_TYPE_ALL,
            nargs="?",
            help=f"Only process incoming or outgoing pending webmention content. Default: {PENDING_TYPE_ALL}",
        )

    def handle(self, pending_type, *args, **options):
        incoming = pending_type == PENDING_TYPE_INCOMING
        outgoing = pending_type == PENDING_TYPE_OUTGOING

        if pending_type == PENDING_TYPE_ALL:
            incoming = True
            outgoing = True

        self.stdout.write(f"Checking for pending webmentions [{pending_type}]...")

        handle_pending_webmentions(incoming=incoming, outgoing=outgoing)



# Generated by Django 4.1 on 2022-08-27 16:47

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("mentions", "0007_dashboardpermissionproxy"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="webmention",
            options={"ordering": ["-created_at"]},
        ),
    ]



from typing import Callable, Optional, Type

from django.apps import apps
from django.urls import ResolverMatch

from mentions import config, contract, options
from mentions.exceptions import BadUrlConfig, NoModelForUrlPath, OptionalDependency
from mentions.helpers.resolution import get_model_for_url_by_helper
from mentions.helpers.thirdparty.wagtail.proxy import RoutablePageMixin
from mentions.helpers.thirdparty.wagtail.util import get_annotation_from_viewfunc
from mentions.helpers.types import MentionableImpl, ModelClass
from mentions.models.mixins import MentionableMixin

__all__ = [
    "get_model_for_url_by_wagtail",
    "autopage_page_resolver",
]


def get_model_for_url_by_wagtail(match: ResolverMatch) -> MentionableMixin:
    """Try to resolve a Wagtail Page instance, if Wagtail is installed.

    If using RoutablePageMixin you must replace the Wagtail @path/@re_path
    decorators with @mentions_wagtail_path/@mentions_wagtail_re_path to add
    the metadata required to resolve the correct target Page.
    """

    if not config.is_wagtail_installed():
        raise OptionalDependency("wagtail")

    import wagtail.views

    if match.func != wagtail.views.serve:
        raise OptionalDependency("wagtail")

    from wagtail.models.sites import get_site_for_hostname

    site = get_site_for_hostname(options.domain_name(), None)
    path = match.args[0]
    path_components = [component for component in path.split("/") if component]

    page, args, kwargs = site.root_page.localized.specific.route(None, path_components)

    if isinstance(page, MentionableMixin):
        return page

    if not isinstance(page, RoutablePageMixin):
        raise NoModelForUrlPath()

    view_func, view_args, view_kwargs = args

    kwarg_mapping = get_annotation_from_viewfunc(view_func)
    view_kwargs.update(kwarg_mapping)

    model_name = view_kwargs.get(contract.URLPATTERNS_MODEL_NAME)

    try:
        model_class: Type[MentionableMixin] = resolve_model(model_name, page)

    except LookupError:
        raise BadUrlConfig(f"Cannot find model `{model_name}`!")

    return get_model_for_url_by_helper(model_class, view_args, view_kwargs)


def autopage_page_resolver(
    model_class: ModelClass,
    lookup: dict,
    view_func: Callable,
) -> Callable:
    def wrapped_view_func(self, request, *args, **kwargs):
        resolved_model = resolve_model(model_class, view_func)
        lookup_kwargs = {**lookup, **kwargs}
        page = get_model_for_url_by_helper(resolved_model, args, lookup_kwargs)

        return view_func(self, request, page)

    return wrapped_view_func


def resolve_app_name(module_name: str) -> Optional[str]:
    app_configs = apps.get_app_configs()
    for conf in app_configs:
        if module_name.startswith(conf.name):
            return conf.label

    raise LookupError(f"Cannot find app for module {module_name}")


def resolve_model(model_class: ModelClass, context: object) -> Type[MentionableImpl]:
    """Resolve a model type from an identifier.

    model_class may be:
    - a class already.
    - a dotted string name for a class.
    - a simple class name. In this case, use context.__module__ to try and
      construct a valid dotted name for the class.
    """
    if not isinstance(model_class, str):
        return model_class

    if "." in model_class:
        return apps.get_model(model_class)

    app_name = resolve_app_name(context.__module__)
    return apps.get_model(f"{app_name}.{model_class}")



from typing import Optional

from django.conf import settings

from mentions.exceptions import (
    BadUrlConfig,
    NoModelForUrlPath,
    TargetDoesNotExist,
    TargetWrongDomain,
)
from mentions.resolution import get_model_for_url

__all__ = [
    "get_target_object",
]

from mentions.util.url import get_domain


def get_target_object(target_url: str) -> Optional["MentionableMixin"]:
    """Confirm that the page exists on our server and return object.

    Args:
        target_url: Our URL that is being mentioned.

    Raises:
        TargetWrongDomain: If the target_url points to a domain not listed in settings.ALLOWED_HOSTS
    """
    domain = get_domain(target_url)

    if domain not in settings.ALLOWED_HOSTS:
        raise TargetWrongDomain(f"Wrong domain: {domain} (from url={target_url})")

    try:
        return get_model_for_url(target_url)

    except NoModelForUrlPath:
        # target_url is not configured to resolve to a model instance.
        pass

    except BadUrlConfig:
        # target_url is not configured to resolve to a model instance.
        pass

    except TargetDoesNotExist:
        pass



import logging
from typing import List, Type

from django.db import models
from django.utils.translation import gettext_lazy as _

from mentions import options
from mentions.exceptions import ImplementationRequired
from mentions.models.mixins.quotable import QuotableMixin

__all__ = [
    "MentionableMixin",
]

log = logging.getLogger(__name__)


def _outgoing_default():
    return options.allow_outgoing_default()


class MentionableMixin(models.Model):
    class Meta:
        abstract = True

    allow_outgoing_webmentions = models.BooleanField(
        _("allow outgoing webmentions"),
        default=_outgoing_default,
    )

    def get_mentions(self) -> List[QuotableMixin]:
        from mentions.resolution import get_mentions_for_object

        return get_mentions_for_object(self)

    def get_mentions_json(self) -> List[dict]:
        from mentions.views.serialize import serialize_mentions

        return serialize_mentions(self.get_mentions())

    def get_absolute_url(self):
        raise ImplementationRequired(
            f"{self.__class__.__name__} does not implement get_absolute_url()"
        )

    def get_content_html(self) -> str:
        """
        Return all the HTML-formatted text that should be searched when looking
        for outgoing Webmentions. Any <a> tags found in this content will be
        checked for webmention support.

        Typically this will just be the main text of your model, but
        you may also want to include content from any other text fields
        such as a summary or abstract.

        Example:
            def get_content_html(self) -> str:
                return f'{self.introduction} {self.main_content}'
        """

        log.warning(
            "Method `MentionableMixin.all_text()` is deprecated, replaced by "
            "`MentionableMixin.get_content_html()`. Please rename the method "
            f"on your model `{self.__class__.__name__}`."
        )

        return self.all_text()

    @classmethod
    def resolve_from_url_kwargs(
        cls: Type["MentionableMixin"],
        **url_kwargs,
    ) -> "MentionableMixin":
        """Resolve a model instance from the given URL captured values.

        You can skip implementing this if you use one of the `urlpatterns`
        helpers: `mentions_path` or `mentions_re_path`.

        By default, an object is resolved via its id using the object_id kwarg.

        You may override this classmethod to retrieve an object using any
        URL parameters in url_kwargs.

        e.g.
        If your urlpatterns path looks like
            "<int:year>/<slug:slug>/",

        then you might override this method as:
            @classmethod
            def resolve_from_url_kwargs(cls, year: int, slug: str, **url_kwargs):
                return cls.objects.get(date__year=year, slug=slug)
        """
        param_mapping = options.default_url_parameter_mapping()
        query = {
            model_field: url_kwargs.get(url_param)
            for url_param, model_field in param_mapping.items()
        }

        return cls.objects.get(**query)

    def should_process_webmentions(self) -> bool:
        """Return True if this instance should process webmentions when saved."""
        return self.allow_outgoing_webmentions

    def save(self, *args, **kwargs):
        super().save(*args, **kwargs)

        if self.should_process_webmentions():
            from mentions.tasks import handle_outgoing_webmentions

            handle_outgoing_webmentions(
                self.get_absolute_url(), self.get_content_html()
            )

    # Deprecated methods below this point
    def mentions(self) -> List[QuotableMixin]:
        """Deprecated in 4.0: Replaced by `get_mentions()`."""
        log.warning(
            "Method `MentionableMixin.mentions()` is deprecated, replaced by "
            "`MentionableMixin.get_mentions()`."
        )
        return self.get_mentions()

    def mentions_json(self):
        """Deprecated in 4.0: Replaced by `get_mentions_json()`."""
        log.warning(
            "Method `MentionableMixin.mentions_json()` is deprecated, replaced by "
            "`MentionableMixin.get_mentions_json()`."
        )
        return self.get_mentions_json()

    def all_text(self) -> str:
        """Deprecated in 4.0: Replaced by `get_content_html()`.

        Only the name has changed - purpose and signature are the same."""
        raise ImplementationRequired(
            f"{self.__class__.__name__} does not implement get_content_html()"
        )



import logging

from django.utils import timezone

log = logging.getLogger(__name__)


class Status:
    """Keep track of any issues that might need to be checked manually."""

    def __init__(self):
        self.notes = []
        self.ok = True

    def info(self, note: str):
        log.info(note)
        self._append(note, "I")
        return self

    def error(self, note: str):
        self.ok = False
        log.error(note)
        self._append(note, "E")
        return self

    def warning(self, note: str) -> "Status":
        log.warning(note)
        self._append(note, "W")
        return self

    def _append(self, note: str, level: str):
        timestamp = timezone.now().strftime("%Y-%m-%d/%H:%M")
        self.notes.append(f"{level}/{timestamp}: {note}")

    def __str__(self):
        return "\n".join(self.notes)[:1023]



from typing import Optional, Type

from django.db import models
from django.utils.translation import gettext_lazy as _

from mentions.models.base import MentionsBaseModel

__all__ = [
    "HCard",
    "update_or_create_hcard",
]


class HCard(MentionsBaseModel):
    """An h-card represents information about a person or organisation.

    Spec: https://microformats.org/wiki/h-card
    """

    name = models.CharField(
        _("name"),
        max_length=50,
        blank=True,
        null=True,
        help_text=_("Name of the person/organisation."),
    )
    avatar = models.URLField(
        _("avatar URL"),
        blank=True,
        null=True,
        help_text=_("Link to their profile image."),
    )
    homepage = models.URLField(
        _("homepage URL"),
        blank=True,
        null=True,
        help_text=_("Link to their homepage."),
    )
    json = models.TextField(
        _("JSON"),
        blank=True,
        help_text=_("Raw json representation of this h-card."),
    )

    def __str__(self):
        return f'HCard(name="{self.name}", avatar="{self.avatar}", homepage="{self.homepage})"'

    class Meta:
        verbose_name = _("h-card")
        verbose_name_plural = _("h-cards")


def update_or_create_hcard(
    homepage: Optional[str],
    name: Optional[str],
    avatar: Optional[str],
    data: str,
) -> HCard:
    """Any individual field may be used to create/retrieve an HCard.

    Ideally, homepage and name are used together.

    Otherwise, the order of precedence is [homepage, name, avatar].
    """

    if homepage and name:
        return _update_first_or_create(
            HCard,
            homepage=homepage,
            name=name,
            defaults={
                "avatar": avatar,
                "json": data,
            },
        )

    if homepage:
        return _update_first_or_create(
            HCard,
            homepage=homepage,
            name=None,
            defaults={
                "avatar": avatar,
                "json": data,
            },
        )

    if name:
        return _update_first_or_create(
            HCard,
            homepage=None,
            name=name,
            defaults={
                "avatar": avatar,
                "json": data,
            },
        )

    if avatar:
        return _update_first_or_create(
            HCard,
            homepage=None,
            name=None,
            avatar=avatar,
            defaults={
                "json": data,
            },
        )


def _update_first_or_create(
    model_cls: Type[MentionsBaseModel],
    defaults: dict,
    **query,
):
    try:
        return model_cls.objects.update_or_create(
            **query,
            defaults=defaults,
        )[0]

    except model_cls.MultipleObjectsReturned:
        instance = model_cls.objects.filter(**query).first()
        for key, value in defaults.items():
            setattr(instance, key, value)

        instance.save()
        return instance



from django import template

from mentions.templatetags.webmentions import webmentions_endpoint

__all__ = [
    "webmentions_endpoint",
]

register = template.Library()


@register.simple_tag
def webmention_endpoint():
    """Deprecated: kept available here to avoid breaking change (for now)."""
    return webmentions_endpoint()



from .hcard import HCard
from .outgoing_status import OutgoingWebmentionStatus
from .pending import PendingIncomingWebmention, PendingOutgoingContent
from .proxy import DashboardPermissionProxy
from .simple_mention import SimpleMention
from .webmention import Webmention



from django.db import models
from django.utils.translation import gettext_lazy as _

from mentions.models.base import MentionsBaseModel
from mentions.models.mixins import RetryableMixin

__all__ = [
    "OutgoingWebmentionStatus",
    "get_or_create_outgoing_webmention",
]


class OutgoingWebmentionStatus(RetryableMixin, MentionsBaseModel):
    """Status tracker for webmentions that you (attempt to) send from your server.

    Used primarily for logging of outgoing mentions.
    """

    source_url = models.URLField(
        _("source URL"),
        help_text=_("The URL on your server where this mention originates."),
    )
    target_url = models.URLField(
        _("target URL"),
        help_text=_("The URL that you mentioned."),
    )
    target_webmention_endpoint = models.URLField(
        _("target webmention endpoint"),
        null=True,
        blank=True,
        help_text=_("The endpoint URL to which we sent the webmention."),
    )
    status_message = models.CharField(
        _("status message"),
        max_length=1024,
        help_text=_("Success, or an explanation of what went wrong."),
    )
    response_code = models.PositiveIntegerField(
        _("response code"),
        default=0,
        help_text=_("HTTP response code of the latest attempted submission."),
    )

    successful = models.BooleanField(
        _("successful"),
        default=False,
    )

    def __str__(self):
        return (
            f"[{'OK' if self.successful else 'Failed'}:{self.response_code}] "
            f"{self.source_url} -> {self.target_url}"
        )

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("outgoing webmention")
        verbose_name_plural = _("outgoing webmentions")


def get_or_create_outgoing_webmention(
    source_urlpath: str,
    target_url: str,
    reset_retries: bool = False,
) -> OutgoingWebmentionStatus:
    """Safely get or create a OutgoingWebmentionStatus instance for given URLs.

    Since version 3.0.0 the URLs are effectively treated as 'unique together',
    but there may exist 'duplicate' instances from a previous installation."""

    kwargs = {
        "source_url": source_urlpath,
        "target_url": target_url,
    }
    try:
        status = OutgoingWebmentionStatus.objects.get_or_create(**kwargs)[0]
    except OutgoingWebmentionStatus.MultipleObjectsReturned:
        status = OutgoingWebmentionStatus.objects.filter(**kwargs).first()

    if reset_retries:
        status.reset_retries()

    return status



import logging

from django import template
from django.urls import reverse
from django.utils.html import format_html

from mentions import options, permissions
from mentions.views import view_names

__all__ = [
    "webmentions_dashboard",
    "webmentions_endpoint",
]

log = logging.getLogger(__name__)
register = template.Library()


@register.simple_tag(takes_context=True)
def webmentions_dashboard(context: dict, link_text: str = "Webmentions Dashboard"):
    """Add a customisable link to your webmentions dashboard in a template."""

    has_permission = context["perms"]["mentions"][
        permissions.can_view_dashboard.codename
    ]
    if not has_permission and not options.dashboard_public():
        return ""

    url = reverse(view_names.webmention_dashboard)
    return format_html(
        """<a href="{url}">{link_text}</a>""",
        url=url,
        link_text=link_text,
    )


@register.simple_tag
def webmentions_endpoint():
    """Add your webmentions endpoint <link> to your template <head>.

    e.g.
    {% load webmentions %}
    ...
    <head>
        {% webmentions_endpoint %}
    </head>
    ...
    """
    endpoint = reverse(view_names.webmention_api_incoming)
    return format_html(
        """<link rel="webmention" href="{endpoint}" />""",
        endpoint=endpoint,
    )



from mentions.models.outgoing_status import get_or_create_outgoing_webmention
from mentions.tasks.celeryproxy import get_logger, shared_task
from mentions.tasks.outgoing.local import get_target_links_in_html
from mentions.tasks.outgoing.remote import try_send_webmention

log = get_logger(__name__)

__all__ = [
    "process_outgoing_webmentions",
]


@shared_task
def process_outgoing_webmentions(source_urlpath: str, text: str) -> int:
    """Find links and in the given text and submit a webmention to any servers that support them.

    Spec:
    https://www.w3.org/TR/webmention/#sender-discovers-receiver-webmention-endpoint

    For each link found in text:
        - Check that the link is alive
        - Search for webmention endpoint in:
            - HTTP response headers
            - HTML <head> element
            - HTML <body> element
        - If an endpoint is found, send a webmention notification

    source_urlpath should be the value returned by model.get_absolute_url() -
    it will be appended to settings.DOMAIN_NAME

    Returns:
         Number of outgoing webmentions that were submitted successfully.
    """

    log.info(f"Checking for mentionable links in text from '{source_urlpath}'...")
    mentions_attempted = 0
    mentions_sent = 0
    links_in_text = get_target_links_in_html(text, source_path=source_urlpath)

    if not links_in_text:
        log.debug("No links found in text.")
        return 0

    for link_url in links_in_text:
        outgoing_webmention = get_or_create_outgoing_webmention(
            source_urlpath,
            link_url,
            reset_retries=True,
        )
        result = try_send_webmention(
            source_urlpath,
            link_url,
            outgoing_status=outgoing_webmention,
        )

        if result is None:
            # No webmention endpoint found
            continue

        mentions_attempted += 1
        if result is True:
            mentions_sent += 1

    if mentions_attempted == 0:
        log.debug(f"No mentionable links found in text.")

    elif mentions_sent == mentions_attempted:
        log.info(f"Successfully sent {mentions_sent} webmentions.")

    else:
        log.warning(
            f"Webmention submission errors: {mentions_sent}/{mentions_attempted} submissions were successful."
        )

    return mentions_sent



import logging
from typing import List, Type

from django.apps import apps
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.exceptions import ObjectDoesNotExist
from django.http import Http404, HttpRequest
from django.urls import Resolver404, ResolverMatch, get_resolver

from mentions import config, contract
from mentions.exceptions import (
    BadUrlConfig,
    NoModelForUrlPath,
    OptionalDependency,
    TargetDoesNotExist,
)
from mentions.helpers.resolution import get_model_for_url_by_helper
from mentions.helpers.thirdparty.wagtail import get_model_for_url_by_wagtail
from mentions.models import SimpleMention, Webmention
from mentions.models.mixins import MentionableMixin, QuotableMixin
from mentions.util.url import get_urlpath

log = logging.getLogger(__name__)


__all__ = [
    "get_mentions_for_object",
    "get_mentions_for_view",
    "get_mentions_for_url",
    "get_model_for_url",
]


def get_urlpattern_match(url_path: str) -> ResolverMatch:
    """Resolves a URL path to the corresponding `urlpatterns` entry.

    Raises:
        TargetDoesNotExist: If a match cannot be found.
    """
    try:

        return get_resolver(settings.ROOT_URLCONF).resolve(url_path)
    except Resolver404:
        raise TargetDoesNotExist(
            f"Cannot find a matching urlpattern entry for path={url_path}"
        )


def get_model_for_url(url: str) -> MentionableMixin:
    """Find the model instance represented by the given URL.

    For this to work:
    - the `urlpatterns` entry that corresponds to the model must include
        a `model_name` kwarg with the model class' dotted path.
    - the model must implement the @classmethod `resolve_from_url_kwargs`

    For implementation details see:
     https://github.com/beatonma/django-wm/wiki/URL-Patterns
     https://github.com/beatonma/django-wm/wiki/Models#mentionablemixin

    e.g.
        path(
            r"articles/<int:article_id>/",
            ArticleView.as_view(),
            kwargs={
                "model_name": "blog.Article",
            },
            name="blog-article"),

    Returns:
        An instance of a model that implements MentionableMixin.

    Raises:
        TargetDoesNotExist: Unable to resolve `url_path` to a `ResolverMatch`,
            or it tries to resolve to a model instance that does not exist.
        NoModelForUrlPath: The ResolverMatch does not resolve to a model instance.
    """
    url_path = get_urlpath(url)
    match = get_urlpattern_match(url_path)
    urlpattern_args = [*match.args]
    urlpattern_kwargs = {**match.kwargs}

    try:
        model_name = urlpattern_kwargs.pop(contract.URLPATTERNS_MODEL_NAME)

    except KeyError:
        try:
            return get_model_for_url_by_wagtail(match)

        except OptionalDependency:
            # Wagtail is not installed
            pass

        except Http404:
            raise TargetDoesNotExist(f"Could not resolve a wagtail page for url {url}")

        except KeyError:
            log.debug(
                f"Wagtail match is not mentionable: {url_path} | args={match.args} | kwargs={match.kwargs}"
            )

        raise NoModelForUrlPath()

    try:
        model_class: Type[MentionableMixin] = apps.get_model(model_name)

    except LookupError:
        raise BadUrlConfig(
            f"Cannot find model `{model_name}` - check your urlpatterns!"
        )

    try:
        return get_model_for_url_by_helper(
            model_class,
            urlpattern_args,
            urlpattern_kwargs,
        )

    except KeyError:
        # URL pattern was not created by helper functions.
        pass

    except ObjectDoesNotExist:
        raise TargetDoesNotExist(
            f"Cannot find instance of model `{model_class}` with kwargs=`{urlpattern_kwargs}`"
        )

    try:
        return model_class.resolve_from_url_kwargs(**urlpattern_kwargs)
    except ObjectDoesNotExist:
        raise TargetDoesNotExist(
            f"Cannot find instance of model `{model_class}` with kwargs=`{urlpattern_kwargs}`"
        )


def get_mentions_for_url(url: str) -> List[QuotableMixin]:
    try:
        obj = get_model_for_url(url)
        return obj.get_mentions()

    except NoModelForUrlPath:
        pass

    if "://" not in url:
        url = config.build_url(url)

    return get_public_mentions(target_url=url)


def get_mentions_for_view(request: HttpRequest) -> List[QuotableMixin]:
    return get_mentions_for_url(request.build_absolute_uri())


def get_mentions_for_object(obj: MentionableMixin) -> List[QuotableMixin]:
    ctype = ContentType.objects.get_for_model(obj.__class__)

    return get_public_mentions(content_type=ctype, object_id=obj.id)


def get_public_mentions(**filter) -> List[QuotableMixin]:
    webmentions = Webmention.objects.filter(
        **filter,
        approved=True,
        validated=True,
    )
    simple_mentions = SimpleMention.objects.filter(**filter)

    return list(webmentions) + list(simple_mentions)



from typing import Type

from django.db.models import Model
from django.http import HttpResponseForbidden
from django.shortcuts import render
from django.views import View

from mentions import options, permissions
from mentions.models import (
    OutgoingWebmentionStatus,
    PendingIncomingWebmention,
    PendingOutgoingContent,
    Webmention,
)

__all__ = [
    "WebmentionDashboardView",
]


class WebmentionDashboardView(View):
    def get(self, request):
        if (
            not permissions.can_view_dashboard.has_perm(request.user)
            and not options.dashboard_public()
        ):
            return HttpResponseForbidden()

        def _sample(model_class: Type[Model]):
            return model_class.objects.all().order_by("-created_at")[:5]

        webmentions = _sample(Webmention)
        pending_incoming = _sample(PendingIncomingWebmention)
        pending_outgoing = _sample(PendingOutgoingContent)
        outgoing_statuses = _sample(OutgoingWebmentionStatus)

        return render(
            request,
            "mentions/webmention-dashboard.html",
            {
                "webmentions": webmentions,
                "pending_incoming": pending_incoming,
                "pending_outgoing": pending_outgoing,
                "outgoing_statuses": outgoing_statuses,
            },
        )



from django import forms
from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from mentions.models import (
    HCard,
    OutgoingWebmentionStatus,
    PendingIncomingWebmention,
    PendingOutgoingContent,
    SimpleMention,
    Webmention,
)

RETRYABLEMIXIN_FIELDS = [
    "is_awaiting_retry",
    "last_retry_attempt",
    "retry_attempt_count",
]


@admin.action(permissions=["change"])
def approve_webmention(modeladmin, request, queryset):
    queryset.update(approved=True)


@admin.action(permissions=["change"])
def disapprove_webmention(modeladmin, request, queryset):
    queryset.update(approved=False)


class BaseAdmin(admin.ModelAdmin):
    save_on_top = True


class ClickableUrlMixin:
    def clickable_source_url(self, obj):
        return clickable_link(obj.source_url)

    clickable_source_url.short_description = _("source URL")

    def clickable_target_url(self, obj):
        return clickable_link(obj.target_url)

    clickable_target_url.short_description = _("target URL")


class TextAreaForm(forms.ModelForm):
    class Meta:
        widgets = {
            "quote": forms.Textarea(attrs={"rows": 3}),
            "notes": forms.Textarea(attrs={"rows": 3}),
        }


@admin.register(SimpleMention)
class QuotableAdmin(BaseAdmin):
    form = TextAreaForm
    date_hierarchy = "published"
    list_display = [
        "source_url",
        "target_url",
        "get_hcard_name",
        "published",
    ]
    list_filter = [
        "post_type",
    ]
    readonly_fields = [
        "target_object",
        "published",
    ]
    search_fields = [
        "quote",
        "source_url",
        "target_url",
        "hcard__name",
        "hcard__homepage",
    ]

    def get_hcard_name(self, obj):
        if obj.hcard:
            return obj.hcard.name

    get_hcard_name.short_description = _("h-card name")


@admin.register(Webmention)
class WebmentionAdmin(ClickableUrlMixin, QuotableAdmin):
    form = TextAreaForm
    actions = [
        approve_webmention,
        disapprove_webmention,
    ]
    list_display = [
        "source_url",
        "target_url",
        "get_hcard_name",
        "validated",
        "approved",
        "published",
        "target_object",
    ]
    list_filter = ["validated", "approved"] + QuotableAdmin.list_filter
    fieldsets = (
        (
            "Remote source",
            {
                "fields": (
                    "clickable_source_url",
                    "sent_by",
                    "hcard",
                    "quote",
                    "post_type",
                ),
            },
        ),
        (
            "Local target",
            {
                "fields": (
                    "clickable_target_url",
                    "content_type",
                    "object_id",
                    "target_object",
                ),
            },
        ),
        (
            "Metadata",
            {
                "fields": (
                    "published",
                    "approved",
                    "validated",
                    "notes",
                ),
            },
        ),
    )
    readonly_fields = QuotableAdmin.readonly_fields + [
        "content_type",
        "object_id",
        "clickable_source_url",
        "clickable_target_url",
        "sent_by",
    ]


@admin.register(OutgoingWebmentionStatus)
class OutgoingWebmentionStatusAdmin(ClickableUrlMixin, BaseAdmin):
    date_hierarchy = "created_at"
    list_display = [
        "source_url",
        "target_url",
        "successful",
        "created_at",
    ]
    list_filter = [
        "successful",
        "is_awaiting_retry",
    ]
    search_fields = [
        "source_url",
        "target_url",
    ]
    exclude = [
        "source_url",
        "target_url",
    ]
    readonly_fields = [
        "created_at",
        "clickable_source_url",
        "clickable_target_url",
        "target_webmention_endpoint",
        "status_message",
        "response_code",
        "successful",
        *RETRYABLEMIXIN_FIELDS,
    ]


@admin.register(HCard)
class HCardAdmin(BaseAdmin):
    list_display = ["name", "homepage"]
    search_fields = ["name", "homepage"]


@admin.register(PendingIncomingWebmention)
class PendingIncomingAdmin(BaseAdmin):
    list_filter = [
        "is_awaiting_retry",
    ]
    readonly_fields = [
        "created_at",
        "source_url",
        "target_url",
        "sent_by",
        *RETRYABLEMIXIN_FIELDS,
    ]
    search_fields = [
        "source_url",
        "target_url",
    ]


@admin.register(PendingOutgoingContent)
class PendingOutgoingAdmin(BaseAdmin):
    readonly_fields = [
        "created_at",
        "absolute_url",
        "text",
    ]
    search_fields = [
        "absolute_url",
        "text",
    ]


def clickable_link(url: str) -> str:
    return format_html(f"<a href={url}>{url}</a>")



from typing import Dict, Set

from django.urls import URLPattern

from mentions import contract
from mentions.helpers.types import ModelClass, SharedFieldName, UrlKwarg

__all__ = [
    "get_captured_filters",
    "get_dotted_model_name",
]


def get_dotted_model_name(model_class: ModelClass) -> str:
    if isinstance(model_class, str):
        return model_class

    model_app = model_class._meta.app_label
    model_name = model_class.__name__

    return f"{model_app}.{model_name}"


def get_captured_filters(
    urlpattern: URLPattern,
) -> Dict[UrlKwarg, Set[SharedFieldName]]:
    """Create default model_filter_map from captured parameters."""
    default_mapping = urlpattern.pattern.converters.keys()
    return {contract.URLPATTERNS_MODEL_FILTER_MAP: set(default_mapping)}



import logging
from typing import Set
from urllib.parse import urljoin

from django.core.exceptions import ValidationError

from mentions import config, options
from mentions.util import find_links_in_html, get_domain, get_url_validator

__all__ = [
    "get_target_links_in_html",
    "is_valid_target",
]


log = logging.getLogger(__name__)


def get_target_links_in_html(html: str, source_path: str) -> Set[str]:
    """Get any links from `html` that should be treated as webmention targets.

    If `options.allow_self_mentions()` is `True`, links that use relative paths
    will be resolved to an absolute URL using `source_path` and
    `options.domain_name()`.

    Args:
        html: HTML-formatted text.
        source_path: The URL path of the HTML source, used to resolve absolute
                     URLs from any links that use relative paths.

    Returns:
        Absolute URLs for any valid links from `html`.
    """
    valid_links = set()
    allow_self_mentions = options.allow_self_mentions()

    for link in find_links_in_html(html):
        if is_valid_target(link, allow_self_mention=allow_self_mentions):
            valid_links.add(link)
            continue

        if "://" in link:
            # Ignore any unsupported schemes.
            continue

        if link.startswith("#"):
            # Ignore local #anchors
            continue

        if allow_self_mentions:
            abs_url = _path_to_absolute_url(link, source_path)
            if is_valid_target(abs_url, allow_self_mention=allow_self_mentions):
                valid_links.add(abs_url)
            else:
                log.warning(f"Constructed absolute url is invalid: {abs_url}")

    return valid_links


def is_valid_target(url: str, allow_self_mention: bool) -> bool:
    """

    Args:
        url: The URL to be tested
        allow_self_mention: Whether this server should be allowed to send
            webmentions to itself.

    Returns:
        True if `url` is a valid URL and should be treated as a possible
        target for webmentions, otherwise False
    """
    try:
        validator = get_url_validator()
        validator(url)

    except ValidationError:
        return False

    if not allow_self_mention:
        if get_domain(url) == options.domain_name():
            return False

    return True


def _path_to_absolute_url(relative_path: str, source_path: str) -> str:
    return urljoin(urljoin(config.base_url(), source_path), relative_path)



# Generated with Django 4.0.2
# Manually edited to squash model deletion and creation steps in a single file.
# PendingIncomingMWebmention and PendingOutgoingContent models are deleted
# and recreated so we can introduce unique constraints on each. Any existing
# instances of those models will be deleted.


from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mentions", "0004_simplemention_post_type_webmention_post_type_and_more"),
    ]

    operations = [
        migrations.DeleteModel(
            name="PendingIncomingWebmention",
        ),
        migrations.DeleteModel(
            name="PendingOutgoingContent",
        ),
        migrations.CreateModel(
            name="PendingIncomingWebmention",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "source_url",
                    models.URLField(
                        help_text="The URL of the content that mentions your content."
                    ),
                ),
                (
                    "target_url",
                    models.URLField(
                        help_text="The URL of the page on your server that is being mentioned."
                    ),
                ),
                (
                    "sent_by",
                    models.URLField(help_text="The origin of the webmention request."),
                ),
            ],
            options={
                "ordering": ["-created_at"],
            },
        ),
        migrations.CreateModel(
            name="PendingOutgoingContent",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("created_at", models.DateTimeField(auto_now_add=True, null=True)),
                (
                    "absolute_url",
                    models.URLField(
                        help_text="URL on our server where the content can be found.",
                        unique=True,
                    ),
                ),
                (
                    "text",
                    models.TextField(
                        help_text="Text that may contain mentionable links. (retrieved via MentionableMixin.all_text())"
                    ),
                ),
            ],
            options={
                "ordering": ["-created_at"],
            },
        ),
        migrations.AddConstraint(
            model_name="pendingincomingwebmention",
            constraint=models.UniqueConstraint(
                fields=("source_url", "target_url"),
                name="unique_source_url_per_target_url",
            ),
        ),
    ]



H_CARD = "h-card"
H_CITE = "h-cite"
H_ENTRY = "h-entry"
H_FEED = "h-feed"

AUTHOR = "p-author"
URL = "u-url"

# Mention types
BOOKMARK = "u-bookmark-of"
LIKE = "u-like-of"
LISTEN = "u-listen-of"
REPLY = "u-in-reply-to"
REPOST = "u-repost-of"
TRANSLATION = "u-translation-of"
WATCH = "u-watch-of"



"""Exceptions raised by the `mentions` app.

Any other exceptions raised should be caught and/or wrapped by one of
these exception classes."""


class WebmentionsException(Exception):
    """Base class for mentions-specific exceptions."""

    pass


class IncomingWebmentionException(WebmentionsException):
    """Base class for an error that occurs while handling a received webmention."""

    pass


class OutgoingWebmentionException(WebmentionsException):
    """Base class for an error that occurs while trying to send a webmention."""

    pass


class WebmentionsConfigurationException(WebmentionsException):
    """Base class for an error related to `django-wm` setup."""

    pass


class BadUrlConfig(WebmentionsConfigurationException):
    """URL resolution completed but did not include required data."""

    pass


class ImplementationRequired(WebmentionsConfigurationException):
    """A MentionableMixin model has not implemented a required method."""

    pass


class TargetWrongDomain(IncomingWebmentionException):
    """Target URL does not point to any domain in settings.ALLOWED_HOSTS."""

    pass


class TargetDoesNotExist(IncomingWebmentionException):
    """Target URL does not point to an object on our server."""

    pass


class SourceNotAccessible(IncomingWebmentionException):
    """Source URL does not exist, or returns an error code."""

    pass


class SourceDoesNotLink(IncomingWebmentionException):
    """Source URL exists but does not contain link to our content."""

    pass


class RejectedByConfig(IncomingWebmentionException):
    """The mention is being rejected due to the current settings configuration."""

    pass


class NoModelForUrlPath(IncomingWebmentionException):
    """URL resolved to a `urlpattern` which did not include `model_name` kwarg."""

    pass


class TargetNotAccessible(OutgoingWebmentionException):
    """Target URL does not exist, or returns an error code."""

    pass


class NotEnoughData(WebmentionsException):
    """Attempted to build a model instance but did not have enough data to do so."""

    pass


class OptionalDependency(WebmentionsException):
    """Attempted to use an optional dependency which is not installed."""

    pass



# Generated by Django 4.1.1 on 2022-09-29 15:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mentions", "0009_alter_outgoingwebmentionstatus_options"),
    ]

    operations = [
        migrations.AlterField(
            model_name="hcard",
            name="avatar",
            field=models.URLField(
                blank=True, help_text="Link to their profile image", null=True
            ),
        ),
        migrations.AlterField(
            model_name="hcard",
            name="name",
            field=models.CharField(
                blank=True,
                help_text="Name of the person/organisation",
                max_length=50,
                null=True,
            ),
        ),
    ]



from django.urls import path, re_path

from mentions.views import view_names
from mentions.views.dashboard import WebmentionDashboardView
from mentions.views.retrieve import GetMentionsByTypeView, GetMentionsView
from mentions.views.submit import WebmentionView

"""
/webmention/
Configure in root project urls.py
e.g.
    from django.conf import settings
    urlpatterns = [
        ...,
        path(r'webmention/', include('mentions.urls')),
    ]
"""
urlpatterns = [
    path("", WebmentionView.as_view(), name=view_names.webmention_api_incoming),
    re_path(
        r"^get/?$",
        GetMentionsView.as_view(),
        name=view_names.webmention_api_get,
    ),
    re_path(
        r"^get-by-type/?$",
        GetMentionsByTypeView.as_view(),
        name=view_names.webmention_api_get_by_type,
    ),
    path(
        "dashboard/",
        WebmentionDashboardView.as_view(),
        name=view_names.webmention_dashboard,
    ),
]



"""
View names for use in urlpatterns and reverse lookup.
"""

"""API endpoint which receives new webmentions."""
webmention_api_incoming = "webmention_api_incoming"

"""API endpoint for retrieving existing webmentions for some target URL."""
webmention_api_get = "webmention_api_get"

"""API endpoint for retrieving existing webmentions, grouped by type, for some target URL."""
webmention_api_get_by_type = "webmention_api_get-by-type"


webmention_dashboard = "webmention-dashboard"



import requests
from requests import Response

from mentions import options

__all__ = [
    "http_get",
    "http_post",
]

HTTP_TIMEOUT_SECONDS = options.timeout()

HTTP_HEADERS = {
    "User-Agent": options.user_agent(),
}


def http_get(url: str) -> Response:
    return requests.get(
        url,
        headers=HTTP_HEADERS,
        timeout=HTTP_TIMEOUT_SECONDS,
    )


def http_post(url: str, data: dict) -> Response:
    return requests.post(
        url,
        data=data,
        headers=HTTP_HEADERS,
        timeout=HTTP_TIMEOUT_SECONDS,
    )



"""JSON serialization for web endpoints."""

from typing import Dict, Iterable, List, Optional

from mentions.models import HCard, SimpleMention, Webmention
from mentions.models.mixins import IncomingMentionType, QuotableMixin

__all__ = [
    "serialize_hcard",
    "serialize_mention",
    "serialize_mentions",
    "serialize_mentions_by_type",
]

from mentions.views import contract


def serialize_mention(mention: QuotableMixin) -> Dict:
    return {
        contract.HCARD: serialize_hcard(mention.hcard),
        contract.MENTION_QUOTE: mention.quote,
        contract.SOURCE_URL: mention.source_url,
        contract.MENTION_PUBLISHED: mention.published,
        contract.MENTION_TYPE: _typeof(mention),
    }


def serialize_mentions(mentions: Iterable[QuotableMixin]) -> List[Dict]:
    return [serialize_mention(mention) for mention in mentions]


def serialize_mentions_by_type(
    mentions: Iterable[QuotableMixin],
) -> Dict[str, List[Dict]]:
    type_names = IncomingMentionType.serialized_names() + [
        contract.MENTION_TYPE_DEFAULT,
        contract.MENTION_TYPE_SIMPLE,
    ]
    types = {name: [] for name in type_names}

    for mention in mentions:
        types[_typeof(mention)].append(serialize_mention(mention))

    return types


def serialize_hcard(hcard: Optional[HCard]) -> Optional[Dict]:
    if hcard is None:
        return None

    return {
        contract.HCARD_NAME: hcard.name,
        contract.HCARD_AVATAR: hcard.avatar,
        contract.HCARD_HOMEPAGE: hcard.homepage,
    }


def _typeof(mention) -> str:
    if isinstance(mention, Webmention):
        return mention.post_type or contract.MENTION_TYPE_DEFAULT

    if isinstance(mention, SimpleMention):
        return contract.MENTION_TYPE_SIMPLE

    raise ValueError(f"Unhandled mention type: {mention}")



# Generated by Django 4.1 on 2022-08-25 14:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        (
            "mentions",
            "0006_add_retryablemixin_pendingincomingwebmention_outgoingwebmentionstatus",
        ),
    ]

    operations = [
        migrations.CreateModel(
            name="DashboardPermissionProxy",
            fields=[
                (
                    "id",
                    models.BigAutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
            ],
            options={
                "permissions": (
                    (
                        "view_webmention_dashboard",
                        "Can view the webmention dashboard/status page.",
                    ),
                ),
                "managed": False,
                "default_permissions": (),
            },
        ),
    ]



import logging
from typing import Optional

from django.http import HttpResponseBadRequest
from django.shortcuts import render
from django.views.decorators.csrf import csrf_exempt
from django.views.generic.base import View

from mentions.forms import SubmitWebmentionForm
from mentions.tasks import handle_incoming_webmention

__all__ = [
    "WebmentionView",
]

log = logging.getLogger(__name__)


# /webmention/
class WebmentionView(View):
    """Handle incoming webmentions."""

    @csrf_exempt
    def dispatch(self, request, *args, **kwargs):
        return super().dispatch(request, *args, **kwargs)

    def get(self, request):
        form = SubmitWebmentionForm()
        return render(
            request,
            "mentions/webmention-submit-manual.html",
            {
                "form": form,
            },
        )

    def post(self, request):
        log.debug("Receiving webmention...")
        form = SubmitWebmentionForm(request.POST)

        if not form.is_valid():
            log.info(f"Received webmention failed form validation: {form.data}")
            return HttpResponseBadRequest()

        data = form.cleaned_data
        source = data["source"]
        target = data["target"]
        client_ip = _get_client_ip(request)

        log.info(
            f"Accepted webmention submission (not yet verified): {source} -> {target}"
        )

        handle_incoming_webmention(source=source, target=target, sent_by=client_ip)
        return render(request, "mentions/webmention-accepted.html", status=202)


def _get_client_ip(request) -> Optional[str]:
    x_forwarded_for = request.META.get("HTTP_X_FORWARDED_FOR")
    if x_forwarded_for:
        ip = x_forwarded_for.split(",")[-1].strip()
    else:
        ip = request.META.get("REMOTE_ADDR")
    return ip



"""
Extra kwargs for any `urlpatterns` entry that is expected to resolve to a
MentionableMixin instance - best left to mentions_path helpers to handle.
"""
URLPATTERNS_MODEL_NAME = "model_name"
URLPATTERNS_MODEL_FILTERS = "model_filters"
URLPATTERNS_MODEL_FILTER_MAP = "model_filter_map"



from dataclasses import dataclass
from typing import Optional
from urllib.parse import urljoin

from mentions.exceptions import SourceDoesNotLink, SourceNotAccessible
from mentions.models import HCard
from mentions.models.mixins import IncomingMentionType
from mentions.tasks.incoming.parsing import (
    find_related_hcard,
    parse_hcard,
    parse_post_type,
)
from mentions.util import html_parser, http_get
from mentions.util.html import find_links_in_soup

__all__ = [
    "get_source_html",
    "get_metadata_from_source",
    "WebmentionMetadata",
]


def get_source_html(source_url: str) -> str:
    """Confirm source exists as HTML and return its content.

    Verify that the source URL returns an HTML page with a successful
    status code.

    Args:
        source_url: The URL that mentions our content.

    Raises:
        SourceNotAccessible: If the `source_url` cannot be resolved, returns an error code, or
                             is an unexpected content type.
    """

    try:
        response = http_get(source_url)
    except Exception as e:
        raise SourceNotAccessible(f"Requests error: {e}")

    if response.status_code >= 300:
        raise SourceNotAccessible(
            f"Source '{source_url}' returned error code [{response.status_code}]"
        )

    content_type = response.headers["content-type"]  # Case-insensitive
    if "text/html" not in content_type:
        raise SourceNotAccessible(
            f"Source '{source_url}' returned unexpected content type: {content_type}"
        )

    return response.text


@dataclass
class WebmentionMetadata:
    post_type: Optional[IncomingMentionType]
    hcard: Optional[HCard]


def get_metadata_from_source(
    html: str,
    target_url: str,
    source_url: str,
) -> WebmentionMetadata:
    """Retrieve contextual data about the mention from the html source.

    Raises:
        SourceDoesNotLink: If the `target_url` is not linked in the given html.
    """

    soup = html_parser(html)
    links = find_links_in_soup(soup)

    for link in links:
        absolute_url = urljoin(source_url, link["href"])
        if absolute_url == target_url:
            break
    else:
        raise SourceDoesNotLink()

    post_type = parse_post_type(link)

    hcard = find_related_hcard(link) or parse_hcard(soup, recursive=False)
    if hcard:
        hcard = coerce_hcard_absolute_urls(hcard, source_url)

    return WebmentionMetadata(
        post_type=post_type.serialized_name() if post_type else None,
        hcard=hcard,
    )


def coerce_hcard_absolute_urls(hcard: HCard, source_url: str) -> HCard:
    """Convert any relative URLs to absolute URLs."""
    updated_fields = []

    if hcard.avatar:
        hcard.avatar = urljoin(source_url, hcard.avatar)
        updated_fields.append("avatar")

    if hcard.homepage:
        hcard.homepage = urljoin(source_url, hcard.homepage)
        updated_fields.append("homepage")

    if updated_fields:
        hcard.save(update_fields=updated_fields)

    return hcard



"""API JSON keys"""
MENTIONS = "mentions"
MENTIONS_BY_TYPE = "mentions_by_type"
MESSAGE = "message"

TARGET_URL = "target_url"
SOURCE_URL = "source_url"

# Mentions
MENTION_QUOTE = "quote"
MENTION_PUBLISHED = "published"
MENTION_TYPE = "type"

MENTION_TYPE_DEFAULT = "webmention"
MENTION_TYPE_SIMPLE = "simple"

# h-card
HCARD = "hcard"
HCARD_NAME = "name"
HCARD_AVATAR = "avatar"
HCARD_HOMEPAGE = "homepage"



import re

from django import template
from django.conf import settings
from django.utils.html import format_html
from django.utils.safestring import mark_safe

register = template.Library()


__all__ = [
    "arrow_icon",
    "more_icon",
    "success_icon",
    "short_url",
]


@register.simple_tag
def success_icon(successful: bool) -> str:
    successful_str = format_html(str(successful).lower())

    if successful:
        title = "Success"
        icon = mark_safe("&check;")

    else:
        title = "Failed"
        icon = mark_safe("&#10539;")

    return format_html(
        """<span class="icon" data-successful="{successful_str}" title="{title}">{icon}</span>""",
        successful_str=successful_str,
        title=title,
        icon=icon,
    )


@register.simple_tag
def short_url(url: str) -> str:
    max_length = 24
    is_self_url = False

    url = re.sub("https?://", "", url)

    if url.startswith("/"):
        is_self_url = True
    else:
        own_hosts = settings.ALLOWED_HOSTS
        for host in own_hosts:
            if url.startswith(host):
                url = url.replace(host, "")
                is_self_url = True
                break

    if len(url) > max_length:
        url = f"{url[:int(max_length / 3)]}…{url[int(max_length * 2 / 3):]}"

    if is_self_url:
        # Replace own domain with 'home' icon for brevity.
        return format_html("""<span class="icon">&#127968;</span>{url}""", url=url)

    return format_html(url)


@register.simple_tag
def arrow_icon() -> str:
    return mark_safe("""<span class="icon">&rarr;</span>""")


@register.simple_tag
def more_icon() -> str:
    return mark_safe(
        """<span class="icon" data-icon="more" title="Click to show detail">&vellip;</span>"""
    )



# Generated by Django 4.1 on 2022-08-18 17:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mentions", "0005_rebuild_pending_models_with_constraints"),
    ]

    operations = [
        migrations.AddField(
            model_name="outgoingwebmentionstatus",
            name="is_awaiting_retry",
            field=models.BooleanField(
                default=True,
                help_text="Whether this object is eligible for another attempt at processing.",
            ),
        ),
        migrations.AddField(
            model_name="outgoingwebmentionstatus",
            name="is_retry_successful",
            field=models.BooleanField(
                default=False,
                editable=False,
                help_text="Whether this object has been processed successfully.",
            ),
        ),
        migrations.AddField(
            model_name="outgoingwebmentionstatus",
            name="last_retry_attempt",
            field=models.DateTimeField(
                blank=True,
                help_text="When we last attempted to process this object.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="outgoingwebmentionstatus",
            name="retry_attempt_count",
            field=models.PositiveSmallIntegerField(
                default=0,
                editable=False,
                help_text="How many times we have attempted to process this object.",
            ),
        ),
        migrations.AddField(
            model_name="pendingincomingwebmention",
            name="is_awaiting_retry",
            field=models.BooleanField(
                default=True,
                help_text="Whether this object is eligible for another attempt at processing.",
            ),
        ),
        migrations.AddField(
            model_name="pendingincomingwebmention",
            name="is_retry_successful",
            field=models.BooleanField(
                default=False,
                editable=False,
                help_text="Whether this object has been processed successfully.",
            ),
        ),
        migrations.AddField(
            model_name="pendingincomingwebmention",
            name="last_retry_attempt",
            field=models.DateTimeField(
                blank=True,
                help_text="When we last attempted to process this object.",
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="pendingincomingwebmention",
            name="retry_attempt_count",
            field=models.PositiveSmallIntegerField(
                default=0,
                editable=False,
                help_text="How many times we have attempted to process this object.",
            ),
        ),
    ]



from django.urls import reverse

from mentions import config
from mentions.views import view_names

__all__ = [
    "WebmentionHeadMiddleware",
]


class WebmentionHeadMiddleware:
    """Automatically add webmention endpoint to HTTP headers of any HttpResponse
    objects created by this server."""

    def __init__(self, get_response):
        self.get_response = get_response

    def __call__(self, request):
        response = self.get_response(request)
        path = reverse(view_names.webmention_api_incoming)

        existing = response.get("Link")
        endpoint_link = f'<{config.build_url(path)}>; rel="webmention"'

        response["Link"] = ",".join([x for x in [existing, endpoint_link] if x])

        return response



# Generated by Django 4.1.3 on 2022-11-23 12:02

from django.db import migrations, models
import django.db.models.deletion
import django.utils.timezone
import mentions.models.webmention


class Migration(migrations.Migration):

    dependencies = [
        ("mentions", "0011_alter_pendingoutgoingcontent_text"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="hcard",
            options={"verbose_name": "h-card", "verbose_name_plural": "h-cards"},
        ),
        migrations.AlterModelOptions(
            name="outgoingwebmentionstatus",
            options={
                "ordering": ["-created_at"],
                "verbose_name": "outgoing webmention",
                "verbose_name_plural": "outgoing webmentions",
            },
        ),
        migrations.AlterModelOptions(
            name="pendingincomingwebmention",
            options={
                "ordering": ["-created_at"],
                "verbose_name": "pending incoming webmention",
                "verbose_name_plural": "pending incoming webmentions",
            },
        ),
        migrations.AlterModelOptions(
            name="pendingoutgoingcontent",
            options={
                "ordering": ["-created_at"],
                "verbose_name": "pending outgoing content",
                "verbose_name_plural": "pending outgoing content",
            },
        ),
        migrations.AlterModelOptions(
            name="webmention",
            options={
                "ordering": ["-created_at"],
                "verbose_name": "webmention",
                "verbose_name_plural": "webmentions",
            },
        ),
        migrations.AlterField(
            model_name="hcard",
            name="avatar",
            field=models.URLField(
                blank=True,
                help_text="Link to their profile image.",
                null=True,
                verbose_name="avatar URL",
            ),
        ),
        migrations.AlterField(
            model_name="hcard",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, null=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="hcard",
            name="homepage",
            field=models.URLField(
                blank=True,
                help_text="Link to their homepage.",
                null=True,
                verbose_name="homepage URL",
            ),
        ),
        migrations.AlterField(
            model_name="hcard",
            name="json",
            field=models.TextField(
                blank=True,
                help_text="Raw json representation of this h-card.",
                verbose_name="JSON",
            ),
        ),
        migrations.AlterField(
            model_name="hcard",
            name="name",
            field=models.CharField(
                blank=True,
                help_text="Name of the person/organisation.",
                max_length=50,
                null=True,
                verbose_name="name",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, null=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="is_awaiting_retry",
            field=models.BooleanField(
                default=True,
                help_text="Whether this object is eligible for another attempt at processing.",
                verbose_name="awaiting retry",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="is_retry_successful",
            field=models.BooleanField(
                default=False,
                editable=False,
                help_text="Whether this object has been processed successfully.",
                verbose_name="retry successful",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="last_retry_attempt",
            field=models.DateTimeField(
                blank=True,
                help_text="When we last attempted to process this object.",
                null=True,
                verbose_name="latest retry attempt",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="response_code",
            field=models.PositiveIntegerField(
                default=0,
                help_text="HTTP response code of the latest attempted submission.",
                verbose_name="response code",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="retry_attempt_count",
            field=models.PositiveSmallIntegerField(
                default=0,
                editable=False,
                help_text="How many times we have attempted to process this object.",
                verbose_name="retry attempt count",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="source_url",
            field=models.URLField(
                help_text="The URL on your server where this mention originates.",
                verbose_name="source URL",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="status_message",
            field=models.CharField(
                help_text="Success, or an explanation of what went wrong.",
                max_length=1024,
                verbose_name="status message",
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="successful",
            field=models.BooleanField(default=False, verbose_name="successful"),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="target_url",
            field=models.URLField(
                help_text="The URL that you mentioned.", verbose_name="target URL"
            ),
        ),
        migrations.AlterField(
            model_name="outgoingwebmentionstatus",
            name="target_webmention_endpoint",
            field=models.URLField(
                blank=True,
                help_text="The endpoint URL to which we sent the webmention.",
                null=True,
                verbose_name="target webmention endpoint",
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, null=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="is_awaiting_retry",
            field=models.BooleanField(
                default=True,
                help_text="Whether this object is eligible for another attempt at processing.",
                verbose_name="awaiting retry",
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="is_retry_successful",
            field=models.BooleanField(
                default=False,
                editable=False,
                help_text="Whether this object has been processed successfully.",
                verbose_name="retry successful",
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="last_retry_attempt",
            field=models.DateTimeField(
                blank=True,
                help_text="When we last attempted to process this object.",
                null=True,
                verbose_name="latest retry attempt",
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="retry_attempt_count",
            field=models.PositiveSmallIntegerField(
                default=0,
                editable=False,
                help_text="How many times we have attempted to process this object.",
                verbose_name="retry attempt count",
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="sent_by",
            field=models.URLField(
                help_text="The origin of the webmention request.",
                verbose_name="sent by",
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="source_url",
            field=models.URLField(
                help_text="The URL of the content that mentions your content.",
                verbose_name="source URL",
            ),
        ),
        migrations.AlterField(
            model_name="pendingincomingwebmention",
            name="target_url",
            field=models.URLField(
                help_text="The URL of the page on your server that is being mentioned.",
                verbose_name="target URL",
            ),
        ),
        migrations.AlterField(
            model_name="pendingoutgoingcontent",
            name="absolute_url",
            field=models.URLField(
                help_text="URL on our server where the content can be found.",
                unique=True,
                verbose_name="absolute URL",
            ),
        ),
        migrations.AlterField(
            model_name="pendingoutgoingcontent",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, null=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="pendingoutgoingcontent",
            name="text",
            field=models.TextField(
                help_text="HTML text that may contain mentionable links.",
                verbose_name="HTML content",
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, null=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="hcard",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="mentions.hcard",
                verbose_name="h-card",
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="object_id",
            field=models.PositiveIntegerField(
                blank=True, null=True, verbose_name="object ID"
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="post_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("bookmark", "Bookmark"),
                    ("like", "Like"),
                    ("listen", "Listen"),
                    ("reply", "Reply"),
                    ("repost", "Repost"),
                    ("translation", "Translation"),
                    ("watch", "Watch"),
                ],
                help_text="Type (e.g. like, reply, etc.) of mention, if specified.",
                max_length=64,
                null=True,
                verbose_name="post type",
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="published",
            field=models.DateTimeField(
                default=django.utils.timezone.now, verbose_name="published"
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="quote",
            field=models.CharField(
                blank=True,
                help_text="A short excerpt from the quoted piece.",
                max_length=300,
                null=True,
                verbose_name="quote",
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="source_url",
            field=models.URLField(
                help_text="The URL that mentions our content.",
                verbose_name="source URL",
            ),
        ),
        migrations.AlterField(
            model_name="simplemention",
            name="target_url",
            field=models.URLField(
                help_text="Our URL that is mentioned.", verbose_name="target URL"
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="approved",
            field=models.BooleanField(
                default=mentions.models.webmention._approve_default,
                help_text="Allow this webmention to appear publicly.",
                verbose_name="approved",
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="created_at",
            field=models.DateTimeField(
                auto_now_add=True, null=True, verbose_name="created at"
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="hcard",
            field=models.ForeignKey(
                blank=True,
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                to="mentions.hcard",
                verbose_name="h-card",
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="notes",
            field=models.CharField(
                blank=True,
                help_text="A description of any errors encountered when building this Webmention.",
                max_length=1024,
                verbose_name="notes",
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="object_id",
            field=models.PositiveIntegerField(
                blank=True, null=True, verbose_name="object ID"
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="post_type",
            field=models.CharField(
                blank=True,
                choices=[
                    ("bookmark", "Bookmark"),
                    ("like", "Like"),
                    ("listen", "Listen"),
                    ("reply", "Reply"),
                    ("repost", "Repost"),
                    ("translation", "Translation"),
                    ("watch", "Watch"),
                ],
                help_text="Type (e.g. like, reply, etc.) of mention, if specified.",
                max_length=64,
                null=True,
                verbose_name="post type",
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="published",
            field=models.DateTimeField(
                default=django.utils.timezone.now, verbose_name="published"
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="quote",
            field=models.CharField(
                blank=True,
                help_text="A short excerpt from the quoted piece.",
                max_length=300,
                null=True,
                verbose_name="quote",
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="sent_by",
            field=models.URLField(
                blank=True,
                help_text="Source address of the HTTP request that sent this webmention.",
                verbose_name="sent by",
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="source_url",
            field=models.URLField(
                help_text="The URL that mentions our content.",
                verbose_name="source URL",
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="target_url",
            field=models.URLField(
                help_text="Our URL that is mentioned.", verbose_name="target URL"
            ),
        ),
        migrations.AlterField(
            model_name="webmention",
            name="validated",
            field=models.BooleanField(
                default=False,
                help_text="True if both source and target have been validated, confirmed to exist, and source really does link to target.",
                verbose_name="validated",
            ),
        ),
    ]



# Generated by Django 4.1.3 on 2022-11-09 18:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("mentions", "0010_alter_hcard_avatar_alter_hcard_name"),
    ]

    operations = [
        migrations.AlterField(
            model_name="pendingoutgoingcontent",
            name="text",
            field=models.TextField(
                help_text="Text that may contain mentionable links. (retrieved via MentionableMixin.get_content_html())"
            ),
        ),
    ]



from typing import Optional

from bs4 import Tag

from mentions import microformats
from mentions.models.mixins import IncomingMentionType

__all__ = [
    "parse_post_type",
]


def parse_post_type(link: Tag) -> Optional[IncomingMentionType]:
    """Return any available type information in the context of the link.

    This may be available as a class on the link itself, or on a parent element
    that is marked with h-cite."""

    def find_mention_type_in_classlist(element: Tag) -> Optional[IncomingMentionType]:
        if element.has_attr("class"):
            classes = set(element["class"])

            for _type in IncomingMentionType.__members__.values():
                if _type.value in classes:
                    return _type

    link_type = find_mention_type_in_classlist(link)
    if link_type is not None:
        return link_type

    hcite = link.find_parent(class_=microformats.H_CITE)
    if hcite:
        return find_mention_type_in_classlist(hcite)



import logging

from mentions import options

__all__ = [
    "get_logger",
    "shared_task",
]

log = logging.getLogger(__name__)


def _noop_shared_task(func, *args, **kwargs):
    """No-op replacement for @decorators that may not be available.

    If the user disables celery via `settings.WEBMENTIONS_USE_CELERY = False`,
    `celery` may not be installed. If `celery` cannot be imported then we need to
    provide an implementation so that un-importable `@shared_task` decorators
    don't break everything.
    """

    class Proxy:
        def delay(self, *args, **kwargs):
            raise NotImplementedError(
                "Called delay() on shared_task but `celery` is not installed! "
                "To disable Celery in `django-wm`, make sure "
                "settings.MENTION_USE_CELERY is False."
            )

        def __call__(self, *args, **kwargs):
            if options.use_celery():
                log.warning("Celery is not installed!")
            return func(*args, **kwargs)

    return (lambda *a, **kw: Proxy())(*args, **kwargs)


try:
    from celery import shared_task
    from celery.utils.log import get_task_logger as _get_task_logger

    get_logger = _get_task_logger

except (ImportError, ModuleNotFoundError):
    shared_task = _noop_shared_task
    get_logger = logging.getLogger



from django.db import models
from django.utils.translation import gettext_lazy as _

__all__ = [
    "MentionsBaseModel",
]


class MentionsBaseModel(models.Model):
    created_at = models.DateTimeField(_("created at"), auto_now_add=True, null=True)

    class Meta:
        abstract = True



import logging
from typing import Callable, Optional, Sequence

from django.urls import URLPattern
from django.urls import path as django_path
from django.urls import re_path as django_re_path

from mentions import contract
from mentions.helpers.thirdparty.wagtail.proxy import wagtail_path, wagtail_re_path
from mentions.helpers.thirdparty.wagtail.resolution import autopage_page_resolver
from mentions.helpers.thirdparty.wagtail.util import annotate_viewfunc
from mentions.helpers.types import ModelClass, ModelFilter, ModelFilterMap
from mentions.helpers.util import get_captured_filters

__all__ = [
    "mentions_wagtail_path",
    "mentions_wagtail_re_path",
    "mentions_wagtail_route",
]

log = logging.getLogger(__name__)


def _path(
    wagtail_path_func: Callable,
    django_path_func: Callable,
    pattern: str,
    model_class: ModelClass,
    model_filters: Optional[Sequence[str]],
    model_filter_map: Optional[ModelFilterMap],
    name: Optional[str],
    autopage: bool,
):
    """Drop-in replacement for the Wagtail routable @path/@re_path decorators.

    Attach model lookup metadata directly to the view func, allowing us to
    resolve the correct Page instance.

    Args:
        wagtail_path_func: wagtail routable @path, @re_path
        django_path_func: django.urls path, re_path
        pattern: A URL pattern [passed to view func].
        model_class: The type of MentionableMixin model that this path represents.
        model_filters: (re_path only) An ordered list of model field names which
            are represented by unnamed groups in the regex pattern.
        model_filter_map: A mapping of captured kwarg names to model field names, if they differ.
            This may be a {captured_name: model_field_name} dictionary, or
            a list of (captured_name, model_field_name) tuples.
        name: Used for reverse lookup [passed to view func]
        autopage: Use with care! Handy but likely controversial.
            If True, automatically resolve the target page and pass it to the view_func.
            - Reduces boilerplate when defining views as you don't have to define
                the same query twice (decorator + view func).
            - The resulting view function will have the signature `def func(self, request, page)`,
              with no additional args or kwargs allowed. This is an intentional restriction to make
              sure this is only used for simple cases.
    """

    def decorator(view_func, *args, **kwargs):
        urlpattern: URLPattern = django_path_func(
            pattern,
            lambda: 0,  # fake View
        )

        lookup = (
            {contract.URLPATTERNS_MODEL_FILTER_MAP: model_filter_map}
            if model_filter_map
            else get_captured_filters(urlpattern)
        )

        if model_filters:
            lookup[contract.URLPATTERNS_MODEL_FILTERS] = model_filters

        wagtail_path = wagtail_path_func(pattern, name=name)

        if autopage:
            view_func = autopage_page_resolver(model_class, lookup, view_func)

        view_func = annotate_viewfunc(view_func, model_class, lookup)
        path = wagtail_path(view_func, *args, **kwargs)

        return path

    return decorator


def mentions_wagtail_path(
    pattern: str,
    model_class: ModelClass,
    model_filter_map: Optional[ModelFilterMap] = None,
    name: str = None,
    autopage=False,
):
    """Drop-in replacement for the Wagtail routable @path decorator.

    Attach model lookup metadata directly to the view func, allowing us to
    resolve the correct Page instance.

    Args:
        pattern: A URL pattern [passed to view func].
        model_class: The type of MentionableMixin model that this path represents.
        model_filter_map: A mapping of captured kwarg names to model field names, if they differ.
            This may be a {captured_name: model_field_name} dictionary, or
            a list of (captured_name, model_field_name) tuples.
        name: Used for reverse lookup [passed to view func]
        autopage: Use with care! Handy but likely controversial.
            Default False.
            If True, automatically resolve the target page and pass it to the view_func.
            - Reduces boilerplate when defining views as you don't have to define
                the same query twice (decorator + view func).
            - The resulting view function will have the signature `def func(self, request, page)`,
              with no additional args or kwargs allowed. This is an intentional restriction to make
              sure this is only used for simple cases.
    """
    return _path(
        wagtail_path_func=wagtail_path,
        django_path_func=django_path,
        pattern=pattern,
        model_class=model_class,
        model_filters=None,
        model_filter_map=model_filter_map,
        name=name,
        autopage=autopage,
    )


def mentions_wagtail_re_path(
    pattern: str,
    model_class: ModelClass,
    model_filters: Optional[Sequence[ModelFilter]] = None,
    model_filter_map: Optional[ModelFilterMap] = None,
    name: str = None,
    autopage=False,
):
    """Drop-in replacement for the Wagtail routable @re_path decorator.

    Attach model lookup metadata directly to the view func, allowing us to
    resolve the correct Page instance.

    As with the core Django equivalent `re_path`, regex may use either named or
    unnamed groups but not both.

    Args:
        pattern: A regex URL pattern [passed to view func].
        model_class: The type of MentionableMixin model that this path represents.
        model_filters: An ordered list of model field names which are represented
            by unnamed groups in the regex pattern.
        model_filter_map: A mapping of captured group names to model field
            names, if they differ. This may be a {captured_name: model_field_name}
            dictionary, or a sequence of (captured_name, model_field_name) tuples.
        name: Used for reverse lookup [passed to view func]
        autopage: Use with care! Handy but likely controversial.
            Default False.
            If True, automatically resolve the target page and pass it to the view_func.
            - Reduces boilerplate when defining views as you don't have to define
                the same query twice (decorator + view func).
            - The resulting view function will have the signature `def func(self, request, page)`,
              with no additional args or kwargs allowed. This is an intentional restriction to make
              sure this is only used for simple cases.
    """
    return _path(
        wagtail_path_func=wagtail_re_path,
        django_path_func=django_re_path,
        pattern=pattern,
        model_class=model_class,
        model_filters=model_filters,
        model_filter_map=model_filter_map,
        name=name,
        autopage=autopage,
    )


mentions_wagtail_route = mentions_wagtail_re_path



from django.db import models
from django.utils.translation import gettext_lazy as _

from mentions import options
from mentions.models.base import MentionsBaseModel
from mentions.models.mixins import QuotableMixin

__all__ = [
    "Webmention",
]


def _approve_default():
    return options.auto_approve()


class Webmention(QuotableMixin, MentionsBaseModel):
    """An incoming webmention that is received by your server."""

    sent_by = models.URLField(
        _("sent by"),
        blank=True,
        help_text=_("Source address of the HTTP request that sent this webmention."),
    )

    approved = models.BooleanField(
        _("approved"),
        default=_approve_default,
        help_text=_("Allow this webmention to appear publicly."),
    )
    validated = models.BooleanField(
        _("validated"),
        default=False,
        help_text=_(
            "True if both source and target have been validated, "
            "confirmed to exist, and source really does link to target."
        ),
    )

    notes = models.CharField(
        _("notes"),
        max_length=1024,
        blank=True,
        help_text=_(
            "A description of any errors encountered when building this Webmention."
        ),
    )

    def approve(self):
        self.approved = True

    class Meta:
        ordering = ["-created_at"]
        verbose_name = _("webmention")
        verbose_name_plural = _("webmentions")

    def __str__(self):
        return (
            f"{self.source_url} -> {self.target_url} "
            f"[validated={self.validated}, approved={self.approved},"
            f"content_type={self.content_type}, id={self.object_id}]"
        )



import logging

from django.core.exceptions import BadRequest
from django.http import JsonResponse
from django.views import View

from mentions import config
from mentions.exceptions import TargetDoesNotExist
from mentions.resolution import get_mentions_for_url
from mentions.views import contract
from mentions.views.serialize import serialize_mentions, serialize_mentions_by_type

__all__ = [
    "GetMentionsView",
    "GetMentionsByTypeView",
]

log = logging.getLogger(__name__)


def build_url(request) -> str:
    for_urlpath = request.GET.get("url")

    if not for_urlpath:
        raise BadRequest()

    return config.build_url(for_urlpath)


# /webmention/get
class GetMentionsView(View):
    def get(self, request):
        """Return any mentions associated with a given url."""
        full_target_url = build_url(request)

        try:
            mentions = get_mentions_for_url(full_target_url)

            return JsonResponse(
                {
                    contract.TARGET_URL: full_target_url,
                    contract.MENTIONS: serialize_mentions(mentions),
                }
            )

        except TargetDoesNotExist as e:
            log.warning(e)
            return JsonResponse(
                {
                    contract.TARGET_URL: full_target_url,
                    contract.MESSAGE: "Target not found",
                    contract.MENTIONS: [],
                },
                status=404,
            )


# /webmention/get_by_type
class GetMentionsByTypeView(View):
    def get(self, request):
        """Return any mentions associated with a given url, grouped by type."""
        full_target_url = build_url(request)

        try:
            mentions = get_mentions_for_url(full_target_url)

            return JsonResponse(
                {
                    contract.TARGET_URL: full_target_url,
                    contract.MENTIONS_BY_TYPE: serialize_mentions_by_type(mentions),
                }
            )

        except TargetDoesNotExist as e:
            log.warning(e)
            return JsonResponse(
                {
                    contract.TARGET_URL: full_target_url,
                    contract.MESSAGE: "Target not found",
                    contract.MENTIONS_BY_TYPE: serialize_mentions_by_type([]),
                },
                status=404,
            )



"""Helper functions that derive from `mentions.options` values."""
from urllib.parse import urljoin

from mentions import options


def base_url() -> str:
    """Get the base URL of this server from current `options`."""
    return f"{options.url_scheme()}://{options.domain_name()}"


def build_url(path: str) -> str:
    """Construct an absolute URL for the given path on this server."""
    return urljoin(base_url(), path)


def is_wagtail_installed() -> bool:
    from importlib.util import find_spec

    wagtail_spec = find_spec("wagtail")
    return wagtail_spec is not None



from typing import Callable

from mentions import contract
from mentions.helpers.types import ModelClass
from mentions.helpers.util import get_dotted_model_name

MENTIONS_KWARGS = "_mentions_kwargs"


def annotate_viewfunc(
    view_func: Callable,
    model_class: ModelClass,
    lookup: dict,
) -> Callable:
    setattr(
        view_func,
        MENTIONS_KWARGS,
        {
            contract.URLPATTERNS_MODEL_NAME: get_dotted_model_name(model_class),
            **lookup,
        },
    )
    return view_func


def get_annotation_from_viewfunc(view_func: Callable) -> dict:
    return getattr(view_func, MENTIONS_KWARGS, {})



from typing import Callable, Dict, Optional, Sequence

from django.urls import URLPattern, path, re_path

from mentions import contract
from mentions.helpers.types import ModelClass, ModelFilter, ModelFilterMap
from mentions.helpers.util import get_captured_filters, get_dotted_model_name

__all__ = [
    "mentions_path",
    "mentions_re_path",
]


def _path(
    func: Callable,
    route: str,
    view: Callable,
    model_class: ModelClass,
    model_filters: Optional[Sequence[ModelFilter]],
    model_filter_map: Optional[ModelFilterMap],
    kwargs: Optional[Dict],
    name: Optional[str],
) -> URLPattern:
    """Proxy for `django.urls.path` and `django.urls.re_path` which enables simpler model resolution.

    This will add extra entries to given `kwargs` which enables model resolution
    via `get_model_for_url_by_helper`.

    Removes the need to implement `resolve_from_url_kwargs` on `model_class`.

    Args:
        func: Either `django.urls.path` or `django.urls.re_path`.
        route: URL pattern [passed to func].
        view: A view function or the result of View.as_view() [passed to func].
        model_class: The type of MentionableMixin model that this path represents.
        model_filters: (re_path only) An ordered list of model field names which
            are represented by unnamed groups in the regex pattern.
        model_filter_map: A mapping of captured kwarg names to model field names, if they differ.
            This may be a {captured_name: model_field_name} dictionary, or
            a list of (captured_name, model_field_name) tuples.
        kwargs: extras options for `view` [modified and passed to func]
        name: Used for reverse lookup [passed to func]

    Returns:
        A URLPattern with extra kwargs that enable resolving an instance of
        `model_class` from arguments captured from `route`.
    """
    urlpattern: URLPattern = func(
        route,
        view,
        kwargs=build_model_kwargs(
            model_class,
            model_filters=model_filters,
            model_filter_map=model_filter_map,
            kwargs=kwargs,
        ),
        name=name,
    )

    if model_filter_map is None:
        # Use captured params for object resolution if no custom mapping provided.
        urlpattern.default_args.update(get_captured_filters(urlpattern))

    return urlpattern


def mentions_path(
    route: str,
    view: Callable,
    model_class: ModelClass,
    model_filter_map: Optional[ModelFilterMap] = None,
    kwargs: Optional[Dict] = None,
    name: Optional[str] = None,
):
    """Proxy for `django.urls.path` to enable MentionableMixin model resolution.

    This will add extra entries to given `kwargs` which enables model resolution
    via `get_model_for_url_by_helper`.

    Removes the need to implement `resolve_from_url_kwargs` on `model_class`.

    Args:
        route: URL pattern.
        view: A view function or the result of View.as_view() [passed to func].
        model_class: The type of MentionableMixin model that this path represents.
        model_filter_map: A mapping of captured kwarg names to model field names, if they differ.
            This may be a {captured_name: model_field_name} dictionary, or
            a list of (captured_name, model_field_name) tuples.
        kwargs: extras options for `view`.
        name: Used for reverse lookup.

    Returns:
        A URLPattern with extra kwargs that enable resolving an instance of
        `model_class` from arguments captured from `route`.
    """
    return _path(
        path,
        route=route,
        view=view,
        model_class=model_class,
        model_filters=None,
        model_filter_map=model_filter_map,
        kwargs=kwargs,
        name=name,
    )


def mentions_re_path(
    route: str,
    view: Callable,
    model_class: ModelClass,
    model_filters: Optional[Sequence[ModelFilter]] = None,
    model_filter_map: Optional[ModelFilterMap] = None,
    kwargs: Optional[Dict] = None,
    name: Optional[str] = None,
):
    """Proxy for `django.urls.re_path` to enable MentionableMixin model resolution.

    This will add extra entries to given `kwargs` which enables model resolution
    via `get_model_for_url_by_helper`.

    Removes the need to implement `resolve_from_url_kwargs` on `model_class`.

    Args:
        route: URL pattern.
        view: A view function or the result of View.as_view() [passed to func].
        model_class: The type of MentionableMixin model that this path represents.
        model_filters: An ordered list of model field names which
            are represented by unnamed groups in the regex pattern.
        model_filter_map: A mapping of captured kwarg names to model field names, if they differ.
            This may be a {captured_name: model_field_name} dictionary, or
            a list of (captured_name, model_field_name) tuples.
        kwargs: extras options for `view`.
        name: Used for reverse lookup.

    Returns:
        A URLPattern with extra kwargs that enable resolving an instance of
        `model_class` from arguments captured from `route`.
    """
    return _path(
        re_path,
        route=route,
        view=view,
        model_class=model_class,
        model_filters=model_filters,
        model_filter_map=model_filter_map,
        kwargs=kwargs,
        name=name,
    )


def build_model_kwargs(
    model_class: ModelClass,
    model_filters: Optional[Sequence[ModelFilter]],
    model_filter_map: Optional[Dict],
    kwargs: Optional[Dict],
) -> Dict:
    """Merge kwargs"""
    fields = (
        {contract.URLPATTERNS_MODEL_FILTERS: model_filters} if model_filters else {}
    )

    return {
        **(kwargs or {}),
        contract.URLPATTERNS_MODEL_NAME: get_dotted_model_name(model_class),
        contract.URLPATTERNS_MODEL_FILTER_MAP: model_filter_map,
        **fields,
    }



from typing import Optional, Tuple, Union

from mentions import options
from mentions.exceptions import (
    RejectedByConfig,
    SourceDoesNotLink,
    SourceNotAccessible,
    TargetDoesNotExist,
    TargetWrongDomain,
)
from mentions.models import PendingIncomingWebmention, Webmention
from mentions.models.mixins import MentionableMixin
from mentions.tasks.celeryproxy import get_logger, shared_task
from mentions.tasks.incoming.local import get_target_object
from mentions.tasks.incoming.remote import (
    WebmentionMetadata,
    get_metadata_from_source,
    get_source_html,
)

__all__ = [
    "process_incoming_webmention",
    "verify_webmention",
]

from mentions.tasks.incoming.status import Status

log = get_logger(__name__)


@shared_task
def process_incoming_webmention(source_url: str, target_url: str, sent_by: str) -> None:
    log.info(f"Processing webmention '{source_url}' -> '{target_url}'")

    status = Status()

    try:
        is_verified, target_object, metadata = verify_webmention(
            source_url=source_url, target_url=target_url
        )

    except RejectedByConfig:
        log.warning(
            f"Ignoring received webmention [{source_url} -> {target_url}]: "
            "target does not resolve to a mentionable model instance."
        )
        return

    except SourceNotAccessible:
        _save_for_retry(source_url, target_url, sent_by)
        return

    if not is_verified:
        status.warning(f"Source does not contain a link to '{target_url}'")

    _create_webmention(
        source_url=source_url,
        target_url=target_url,
        sent_by=sent_by,
        target_object=target_object,
        verified=is_verified,
        metadata=metadata,
        notes=status,
    )
    _mark_complete(source_url, target_url)


def verify_webmention(
    source_url: str,
    target_url: str,
) -> Tuple[bool, Optional[MentionableMixin], Optional[WebmentionMetadata]]:
    """If the returned metadata is None, verification"""
    is_verified = False

    try:
        target_object = get_target_object(target_url)

    except TargetWrongDomain:
        log.warning(
            f"Received webmention does not target our domain: {source_url} -> {target_url}"
        )
        raise

    except TargetDoesNotExist:
        target_object = None

    if target_object is None and options.target_requires_model():
        log.warning(
            f"Ignoring received webmention [{source_url} -> {target_url}]: "
            "target does not resolve to a mentionable model instance."
        )
        raise RejectedByConfig(f"No target_object found for url={target_url}")

    try:
        response_html = get_source_html(source_url)

    except SourceNotAccessible:
        raise

    try:
        metadata = get_metadata_from_source(response_html, target_url, source_url)
        is_verified = True
    except SourceDoesNotLink:
        metadata = None

    return is_verified, target_object, metadata


def _create_webmention(
    source_url: str,
    target_url: str,
    sent_by: str,
    verified: bool,
    target_object: Optional[MentionableMixin],
    metadata: Optional[WebmentionMetadata],
    notes: Union[Status, str] = "",
) -> Webmention:
    return Webmention.objects.create(
        source_url=source_url,
        target_url=target_url,
        sent_by=sent_by,
        target_object=target_object,
        validated=verified,
        hcard=metadata.hcard if metadata else None,
        post_type=metadata.post_type if metadata else None,
        notes=str(notes),
    )


def _mark_complete(source_url: str, target_url: str):
    try:
        pending = PendingIncomingWebmention.objects.get(
            target_url=target_url,
            source_url=source_url,
        )
        pending.mark_processing_successful(save=True)
    except PendingIncomingWebmention.DoesNotExist:
        pass


def _save_for_retry(source_url: str, target_url: str, sent_by: str):
    """In case of network errors, create or update PendingIncomingWebmention instance to retry later."""
    pending, _ = PendingIncomingWebmention.objects.get_or_create(
        target_url=target_url,
        source_url=source_url,
        defaults={
            "sent_by": sent_by,
        },
    )

    pending.mark_processing_failed(save=True)



from typing import Dict, Optional, Sequence, Set, Tuple, Type

from mentions import contract
from mentions.helpers.types import (
    MentionableImpl,
    ModelFilter,
    ModelFilterMap,
    UrlKwarg,
)

__all__ = [
    "get_model_for_url_by_helper",
]

TypeSet = Set[Tuple[UrlKwarg, ModelFilter]]


def get_model_for_url_by_helper(
    model_class: Type[MentionableImpl],
    urlpattern_args: Sequence,
    urlpattern_kwargs: Dict,
) -> Optional[MentionableImpl]:
    """Resolve a model instance from urlpattern kwargs, as configured by
    `mentions_path` or `mentions_re_path` helper functions.

    Args:
        model_class: A class that inherits MentionableMixin.
        urlpattern_args: args from a `ResolverMatch`.
        urlpattern_kwargs: kwargs from a `ResolverMatch`.

    Returns:
        An instance of `model_class`.

    Raises:
        KeyError: If the given urlpattern_kwargs does not contain helper values.
    """

    model_filter_map: ModelFilterMap = urlpattern_kwargs.pop(
        contract.URLPATTERNS_MODEL_FILTER_MAP
    )

    mapping = unpack_model_filter_map(model_filter_map)
    query = {value: urlpattern_kwargs[key] for key, value in mapping}

    model_filters: Sequence = urlpattern_kwargs.pop(
        contract.URLPATTERNS_MODEL_FILTERS, []
    )
    for key, value in zip(model_filters, urlpattern_args):
        query[key] = value

    return model_class.objects.get(**query)


def unpack_model_filter_map(mapping: ModelFilterMap) -> TypeSet:
    if isinstance(mapping, dict):
        mapping = set(mapping.items())

    def unpack(item):
        # If mapping is a string, replace it with an 'identity' tuple of it: key == value.
        if isinstance(item, str):
            return item, item
        return item

    mapping = {unpack(x) for x in mapping}

    return mapping



import logging
from itertools import chain

from mentions import options
from mentions.models import (
    OutgoingWebmentionStatus,
    PendingIncomingWebmention,
    PendingOutgoingContent,
)
from mentions.tasks.celeryproxy import shared_task
from mentions.tasks.incoming import process_incoming_webmention
from mentions.tasks.outgoing import (
    is_valid_target,
    process_outgoing_webmentions,
    try_send_webmention,
)

log = logging.getLogger(__name__)


__all__ = [
    "handle_pending_webmentions",
    "handle_incoming_webmention",
    "handle_outgoing_webmentions",
]


@shared_task
def handle_pending_webmentions(incoming: bool = True, outgoing: bool = True):
    """Process any webmentions that are pending processing, including retries.

    Typically run via `manage.py pending_mentions`"""

    if incoming:
        _handle_pending_incoming()

    if outgoing:
        _handle_pending_outgoing()

    if options.use_celery():
        _maybe_reschedule_handle_pending_webmentions()


def handle_incoming_webmention(source: str, target: str, sent_by: str) -> None:
    """Delegate processing to `celery` if available, otherwise store for later.

    If settings.WEBMENTIONS_USE_CELERY is False, create a PendingIncomingWebmention.
    This needs to be processed at some point by running `manage.py pending_mentions`.
    """

    use_celery = options.use_celery()

    if use_celery:
        _task_handle_incoming.delay(source, target, sent_by)

    else:
        PendingIncomingWebmention.objects.get_or_create(
            source_url=source,
            target_url=target,
            defaults={
                "sent_by": sent_by,
            },
        )


@shared_task
def _task_handle_incoming(source: str, target: str, sent_by: str):
    process_incoming_webmention(
        source_url=source,
        target_url=target,
        sent_by=sent_by,
    )
    _maybe_reschedule_handle_pending_webmentions()


def handle_outgoing_webmentions(absolute_url: str, text: str) -> None:
    """Delegate processing to `celery` if available, otherwise store for later.

    If settings.WEBMENTIONS_USE_CELERY is False, create a PendingOutgoingContent.
    This needs to be processed at some point by running `manage.py pending_mentions`.
    """
    use_celery = options.use_celery()

    if use_celery:
        _task_handle_outgoing.delay(absolute_url, text)

    else:
        PendingOutgoingContent.objects.get_or_create(
            absolute_url=absolute_url,
            defaults={
                "text": text,
            },
        )


@shared_task
def _task_handle_outgoing(absolute_url: str, text: str) -> None:
    process_outgoing_webmentions(source_urlpath=absolute_url, text=text)
    _maybe_reschedule_handle_pending_webmentions()


def _handle_pending_incoming():
    for incoming_wm in PendingIncomingWebmention.objects.filter(is_awaiting_retry=True):
        if incoming_wm.can_retry():
            process_incoming_webmention(
                incoming_wm.source_url,
                incoming_wm.target_url,
                incoming_wm.sent_by,
            )

            incoming_wm.refresh_from_db()
            if incoming_wm.is_retry_successful:
                # Webmention created successfully so this is no longer needed.
                incoming_wm.delete()


def _handle_pending_outgoing():
    allow_self_mentions = options.allow_self_mentions()

    for outgoing_retry in OutgoingWebmentionStatus.objects.filter(
        is_awaiting_retry=True,
    ):
        if not is_valid_target(
            outgoing_retry.target_url,
            allow_self_mention=allow_self_mentions,
        ):
            log.warning(f"Target URL is invalid: {outgoing_retry.target_url}")
            outgoing_retry.delete()
            continue

        if outgoing_retry.can_retry():
            try_send_webmention(
                source_urlpath=outgoing_retry.source_url,
                target_url=outgoing_retry.target_url,
                outgoing_status=outgoing_retry,
            )

    for pending_out in PendingOutgoingContent.objects.all():
        process_outgoing_webmentions(pending_out.absolute_url, pending_out.text)
        # OutgoingWebmentionStatus created instead to track status of individual links.
        pending_out.delete()


def _maybe_reschedule_handle_pending_webmentions():
    """Check if there are objects awaiting retry; if so, schedule `handle_pending_webmentions` to run again later."""
    should_reschedule = (
        PendingIncomingWebmention.objects.filter(is_awaiting_retry=True).exists()
        or OutgoingWebmentionStatus.objects.filter(is_awaiting_retry=True).exists()
    )

    if not should_reschedule:
        return

    _reschedule_handle_pending_webmentions()


def _reschedule_handle_pending_webmentions():
    """Using celery, schedule `handle_pending_webmentions` to run again later.

    Only one such task should be scheduled at a time - halt if it is already scheduled."""
    import celery

    # Check if this task is already scheduled.
    task = handle_pending_webmentions
    task_name = ".".join([task.__module__, task.__name__])

    app = celery.current_app
    scheduled = app.control.inspect().scheduled()
    scheduled_task_etas = [
        item["eta"]
        for item in chain.from_iterable(scheduled.values())
        if item["request"]["name"] == task_name
    ]

    if scheduled_task_etas:
        log.info(f"Task '{task_name}' already scheduled: ETA {scheduled_task_etas}")
        return

    # Schedule fresh task
    interval = options.retry_interval()
    task.apply_async(
        countdown=interval,
        expires=interval * 2,
    )
    log.info(f"Scheduled task '{task_name}' in {interval} seconds.")



from django import forms

__all__ = [
    "SubmitWebmentionForm",
]


class SubmitWebmentionForm(forms.Form):
    target = forms.URLField(label="The URL of my page")
    source = forms.URLField(label="The URL of your page")



from django.db import models

from mentions import permissions as perms

__all__ = [
    "DashboardPermissionProxy",
]


class DashboardPermissionProxy(models.Model):
    """Unmanaged model registers permission required to view mentions dashboard."""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (perms.can_view_dashboard.as_tuple(),)



"""A management command to reprocess an existing Webmention."""
from argparse import ArgumentParser
from typing import List, Union

from django.core.management import BaseCommand
from django.db.models import QuerySet

from mentions.models import Webmention
from mentions.tasks.incoming.reverify import reverify_mention


class Command(BaseCommand):
    def add_arguments(self, parser: ArgumentParser):
        parser.add_argument(
            "filters",
            nargs="*",
            help="Space-separated list of field=value queryset filters.",
        )

        parser.add_argument(
            "--all",
            dest="all_mentions",
            action="store_true",
            default=False,
            help="Reverify all webmentions.",
        )

    def handle(self, *args, filters: List[str], all_mentions: bool, **options):
        target_mentions = get_target_mentions(filters, all_mentions)

        changed = []

        for mention in target_mentions:
            if reverify_mention(mention):
                changed.append(mention)

        self.stdout.write(f"Updated {len(changed)} mention(s):")
        for mention in changed:
            self.stdout.write(f"- {mention}")


def get_target_mentions(filters: List[str], all_mentions: bool) -> QuerySet[Webmention]:
    if all_mentions:
        return Webmention.objects.all()

    if not filters:
        raise ValueError(
            "Please provide at least one field=value queryset filter. "
            "To reverify all Webmentions, use --all to explicitly bypass filtering."
        )

    query = {}
    for filter_str in filters:
        filter_, value = filter_str.split("=")
        query[filter_] = parse_filter_value(value)

    return Webmention.objects.filter(**query)


def parse_filter_value(filter_value: str) -> Union[bool, int, float, str]:
    if filter_value == "True":
        return True
    if filter_value == "False":
        return False

    try:
        return int(filter_value)
    except ValueError:
        pass

    try:
        return float(filter_value)
    except ValueError:
        pass

    return filter_value



"""Modules should import wagtail components from here.

The library should work with (and without)wWagtail==3.0.3 and above. The @path
decorator was only introduced in wagtail==4.0.0 so we need to handle it being
unavailable."""

from mentions.exceptions import OptionalDependency

try:
    from wagtail.contrib.routable_page.models import RoutablePageMixin
    from wagtail.contrib.routable_page.models import path as wagtail_path
    from wagtail.contrib.routable_page.models import re_path as wagtail_re_path

except ImportError:

    def config_error(function_name: str, min_wagtail_version: str, django_func):
        """When wagtail is not available or"""

        def fake_decorator(pattern, name):
            def raise_error(*args, **kwargs):
                raise OptionalDependency(
                    f"Tried to use decorator '{function_name}' but "
                    f"wagtail>={min_wagtail_version} is not installed. ("
                    f"pattern={pattern}, name={name})"
                )

            def fake_wrapper(view_func):
                if not hasattr(view_func, "_routablepage_routes"):
                    view_func._routablepage_routes = []

                view_func._routablepage_routes.append(
                    (
                        django_func(
                            pattern,
                            raise_error,
                            name=name or view_func.__name__,
                        ),
                        1000,
                    )
                )
                return view_func

            return fake_wrapper

        return fake_decorator

    try:
        from django.urls import re_path
        from wagtail.contrib.routable_page.models import RoutablePageMixin
        from wagtail.contrib.routable_page.models import route as wagtail_re_path

    except ImportError:
        wagtail_re_path = config_error("mentions_wagtail_re_path", "3.0.3", re_path)

        class RoutablePageMixin:
            pass

    from django.urls import path

    wagtail_path = config_error("mentions_wagtail_path", "4.0", path)



from typing import Set

from bs4 import BeautifulSoup, ResultSet

__all__ = [
    "html_parser",
    "find_links_in_html",
    "find_links_in_soup",
]


def html_parser(content) -> BeautifulSoup:
    return BeautifulSoup(content, features="html5lib")


def find_links_in_html(html: str) -> Set[str]:
    """Get the raw target href of any links in the html."""
    soup = html_parser(html)
    return {a["href"] for a in find_links_in_soup(soup)}


def find_links_in_soup(soup: BeautifulSoup) -> ResultSet:
    return soup.find_all("a", href=True)



# Generated by Django 4.0.2 on 2022-06-30 18:18

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('mentions', '0003_alter_simplemention_published_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='simplemention',
            name='post_type',
            field=models.CharField(blank=True, choices=[('bookmark', 'Bookmark'), ('like', 'Like'), ('listen', 'Listen'), ('reply', 'Reply'), ('repost', 'Repost'), ('translation', 'Translation'), ('watch', 'Watch')], help_text='Type (e.g. like, reply, etc.) of mention, if specified', max_length=64, null=True),
        ),
        migrations.AddField(
            model_name='webmention',
            name='post_type',
            field=models.CharField(blank=True, choices=[('bookmark', 'Bookmark'), ('like', 'Like'), ('listen', 'Listen'), ('reply', 'Reply'), ('repost', 'Repost'), ('translation', 'Translation'), ('watch', 'Watch')], help_text='Type (e.g. like, reply, etc.) of mention, if specified', max_length=64, null=True),
        ),
        migrations.AlterField(
            model_name='webmention',
            name='notes',
            field=models.CharField(blank=True, help_text='A description of any errors encountered when building this Webmention.', max_length=1024),
        ),
    ]



from typing import Dict, Sequence, Tuple, Type, TypeVar, Union

from mentions.models.mixins import MentionableMixin

"""Represents the name of a captured group from a URL pattern."""
UrlKwarg = str

"""Represents the name of a filter that can be used in a Django query.

This may simply be a field name like `title`, or something more complex 
like `date__year`, although Q and F objects are not currently supported."""
ModelFilter = str


"""A name that represents a UrlKwarg AND a ModelFilter.

The same name is used in the URL pattern and the resulting database query."""
SharedFieldName = str

ModelFilterMap = Union[
    Dict[UrlKwarg, ModelFilter],
    Sequence[SharedFieldName],
    Sequence[Tuple[UrlKwarg, ModelFilter]],
]

MentionableImpl = TypeVar("MentionableImpl", bound=MentionableMixin)
ModelClass = Union[Type[MentionableImpl], str]



from dataclasses import dataclass
from typing import Tuple

from django.contrib.auth.models import Permission
from django.utils.translation import gettext_lazy as _

from mentions.apps import MentionsConfig

__all__ = [
    "can_view_dashboard",
]


@dataclass
class MentionsPermission:
    codename: str
    description: str

    def as_tuple(self) -> Tuple[str, str]:
        return self.codename, self.description

    def has_perm(self, user, obj=None) -> bool:
        return user.has_perm(self.fqn(), obj=obj)

    def grant(self, user):
        user.user_permissions.add(self.get_from_db())

    def fqn(self) -> str:
        return f"{MentionsConfig.name}.{self.codename}"

    def get_from_db(self):
        return Permission.objects.get(codename=self.codename)

    def __str__(self):
        return self.fqn()


can_view_dashboard = MentionsPermission(
    "view_webmention_dashboard",
    _("Can view the webmention dashboard/status page."),
)



# Generated by Django 4.0.2 on 2022-03-26 13:13

from django.db import migrations, models
import django.utils.timezone


class Migration(migrations.Migration):

    dependencies = [
        ('mentions', '0002_pendingincomingwebmention_pendingoutgoingcontent'),
    ]

    operations = [
        migrations.AlterField(
            model_name='simplemention',
            name='published',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
        migrations.AlterField(
            model_name='webmention',
            name='published',
            field=models.DateTimeField(default=django.utils.timezone.now),
        ),
    ]
